#!/usr/bin/python

import os
import re
import logging.config
import inspect
import threading
import traceback

#============================ logging =========================================

def filenameCallingModule():
    # fine filename of calling module
    frame,filename,line_number,function_name,lines,index=\
        inspect.getouterframes(inspect.currentframe())[2]
    
    filename = os.path.split(filename)[1]
    filename = filename.split('.')[0]
    
    return filename
    
def setupModuleLogging():
    
    # get the filename of the calling module
    filename = filenameCallingModule()
    
    # create logger with name that filename
    log = logging.getLogger(filename)
    log.setLevel(logging.CRITICAL)
    
    return log

def configureLoggingNoFile(logfilename=None,debugmodules=[]):
    
    #=== create formatter
    
    formatter = logging.Formatter("%(asctime)s [%(name)s:%(levelname)s] %(message)s")
    
    #=== create handlers
    
    # determine the log filename
    if not logfilename:
        logfilename = '{0}.log'.format(filenameCallingModule())
    
    # create a log handler which write to a file
    fileHandler = logging.handlers.RotatingFileHandler(
        logfilename,
        mode            = 'a',
        backupCount     = 5,
        maxBytes        = 2*1024*1024,
    )
    fileHandler.setFormatter(formatter)
    fileHandler.setLevel(logging.INFO)
    
    # create a log handler which print to console
    consoleHandler = logging.StreamHandler()
    consoleHandler.setFormatter(formatter)
    consoleHandler.setLevel(logging.ERROR)
    
    # create a log handler which print to console
    consoleHandler2 = logging.StreamHandler()
    consoleHandler2.setFormatter(logging.Formatter("%(asctime)s %(message)s", "%H:%M:%S"))
    consoleHandler2.setLevel(logging.DEBUG)
    
    #=== associated handlers with loggers
    
    loggerNames = logging.Logger.manager.loggerDict.keys()
    
    for n in loggerNames:
        temp = logging.getLogger(n)
        temp.setLevel(logging.INFO)
        temp.addHandler(fileHandler)
        temp.addHandler(consoleHandler)
    
    for n in debugmodules:
        temp = logging.getLogger(n)
        temp.setLevel(logging.DEBUG)
        temp.addHandler(consoleHandler2)

def configureLogging():
    
    # compute the log filename
    fullpath       = inspect.stack()[1][1]
    filename       = (os.path.split(fullpath)[-1]).split('.')[0]
    logfilename    = '{0}.log'.format(filename)
    
    # find the log configuration file, if any
    logconf = None
    for p in [
            os.path.join('.',  'logging.conf'),
            os.path.join('..', 'logging.conf'),
            os.path.join('app','logging.conf'),
        ]:
        if os.path.exists(p):
            logconf = p
            break
    
    # start logging
    if logconf:
        logging.config.fileConfig(
            fname      = logconf,
            defaults   = {
                'logfilename': logfilename,
            }
        )
    else:
        print "WARNING: no log configuration file could be found."

#============================ formatting ======================================

def handlerThreadCrash(name,err,log):
    output  = []
    output += ['===== crash in thread {0} ====='.format(name)]
    output += ['\nerror:\n']
    output += [str(err)]
    output += ['\ncall stack:\n']
    output += [traceback.format_exc()]
    output  = '\n'.join(output)
    log.critical(output)
    raise

#============================ configuration ===================================

class AppConfig(object):
    '''
    \brief A singleton which contains some configuration, typically for an
        application.
    '''
    
    #===== singleton start
    _instance = None
    _init     = False
    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super(AppConfig, cls).__new__(cls, *args, **kwargs)
        return cls._instance
    #===== singleton stop
    
    def __init__(self):
        
        #===== singleton start
        if self._init:
            return
        self._init = True
        #===== singleton stop
        
        # store params
        self.dataLock   = threading.RLock()
        self.config     = {}
    
    def get(self,*args):
        with self.dataLock:
            if   len(args)==1:
                return self.config.get(args[0])
            elif len(args)==2:
                return self.config.get(args[0],args[1])
            else:
                raise SystemError()
    
    def set(self,name,value):
        with self.dataLock:
            self.config[name] = value
    
    def readFile(self,filename):
        try:
            with open(filename,'r') as f:
                for line in f:
                    line = line.strip()
                    
                    # comments
                    if line.startswith('#'):
                        continue
                    
                    # configuration
                    m = re.search('(\S+)\s*=\s*(\S+)',line)
                    if m:
                        name   = m.group(1)
                        values = m.group(2).split(',')
                        values = [v.strip() for v in values]
                        
                        for i in range(len(values)):
                            # int
                            try:
                                values[i] = int(values[i])
                            except:
                                pass
                            # float
                            try:
                                values[i] = float(values[i])
                            except:
                                pass
                        
                        if len(values)==1:
                            values = values[0]
                        
                        self.set(name,values)
        except IOError:
            # could not open/read file, that's OK
            pass



# ============================ Imports ====================================
import base64
from txp.cloud import pull_current_configuration_from_firestore
from txp.common.json_complex_encoder import ComplexEncoder
import google.cloud.firestore as firestore
from google.cloud import iot_v1
import json
from typing import List, Dict
from txp.cloud import settings
from txp.common.edge.common import PairedEdgesPayload, EdgeDescriptor, MachineMetadata
import sys


# =========================== Constant Data & helper functions ===============================
def _get_edges_from_db(
        db: firestore.Client, configuration_ref, edges_logical_ids: List[str]
):
    """Returns the Edges Documents from the Database that matches the logical ids
    passed in the edges_logical_ids

    Returns:
        The list of found Edges documents in Firestore.
    """
    edges_collection_name = settings.firestore.edges_collection
    configuration_reference_field_path = "configuration_ref"
    edges = (
        db.collection(edges_collection_name)
            .where(configuration_reference_field_path, "==", configuration_ref)
            .where("logical_id", "in", edges_logical_ids)
            .get()
    )
    return list(edges)


def _pull_machine(
        db: firestore.Client, configuration_ref, machine_id: str
):
    """Returns the Machine document associates with the virtual edge actuator
    declared in the logical configuration. """
    machines_collection_name = settings.firestore.machines_collection
    configuration_reference_field_path = "configuration_ref"
    machine_doc = (
        db.collection(machines_collection_name)
        .where(configuration_reference_field_path, "==", configuration_ref)
        .where("machine_id", "==", machine_id)
    ).get()

    if not machine_doc:
        return None

    return machine_doc[0]


def _add_edges_to_machines(
        db: firestore.Client,
        configuration_ref: firestore.DocumentReference,
        machines: List[MachineMetadata],
        edges_refs: Dict[str, firestore.DocumentReference]
):
    for machine in machines:
        machine_doc: firestore.DocumentSnapshot = _pull_machine(db, configuration_ref, machine.machine_id)
        if not machine_doc:
            print(
                f"Unexpected error: Machine document not found for {machine.machine_id}.",
                file=sys.stderr,
            )
            continue

        new_edges_for_machine = []
        for edge in machine.edges_ids:
            if edge in edges_refs:  # If not, then the edge already existed.
                new_edges_for_machine.append(edges_refs[edge])

        old_edges = machine_doc.get("associated_with_edges")
        new_edges = old_edges + new_edges_for_machine
        machine_doc.reference.update({
            "associated_with_edges": new_edges
        })

        print(f"Updated Machine document {machine_doc.id} with logical id {machine.machine_id}\n"
              f"New set of edges: {list(map(lambda edge_ref: edge_ref.get().get('logical_id'), new_edges ))}")


def _create_new_virtual_edges(
        db: firestore.Client, configuration_ref, device_kind_doc_ref,
        edge_descriptors: List[EdgeDescriptor]
):
    """Creates the New virtual Edges in Firestore."""
    edges_references: Dict[str, firestore.DocumentReference] = {}
    for edge_descriptor in edge_descriptors:
        parameters_dict = json.loads(
            json.dumps(edge_descriptor.edge_parameters, cls=ComplexEncoder)
        )
        document_dict = {
            "device_kind": edge_descriptor.device_kind,
            "configuration_ref": configuration_ref,
            "is_of_kind_ref": device_kind_doc_ref,
            "logical_id": edge_descriptor.logical_id,
            "parameters": parameters_dict,
            "type": edge_descriptor.device_type
        }
        edge_ref = db.collection(settings.firestore.edges_collection).add(
            document_dict
        )[1]
        edges_references[edge_descriptor.logical_id] = edge_ref
        print(f"Document created for virtual edge: {edge_descriptor.logical_id}")

    return edges_references


def _create_iot_device(
    project_id, cloud_region, registry_id, device_id
):
    """Create a device to bind to a gateway if it does not exist."""
    client = iot_v1.DeviceManagerClient()
    exists = False
    parent = client.registry_path(project_id, cloud_region, registry_id)
    devices = list(client.list_devices(request={"parent": parent}))

    for device in devices:
        if device.id == device_id:
            exists = True

    # Create the device
    device_template = {
        "id": device_id,
        "gateway_config": {
            "gateway_type": iot_v1.GatewayType.NON_GATEWAY,
            "gateway_auth_method": iot_v1.GatewayAuthMethod.ASSOCIATION_ONLY,
        },
    }

    if not exists:
        res = client.create_device(
            request={"parent": parent, "device": device_template}
        )
        print("Created Device {}".format(res))
    else:
        print("Device exists, skipping")


def _bind_device_to_gateway(
    project_id, cloud_region, registry_id, device_id, gateway_id
):
    """Binds a device to a gateway."""
    client = iot_v1.DeviceManagerClient()

    _create_iot_device(
        project_id, cloud_region, registry_id, device_id
    )

    parent = client.registry_path(project_id, cloud_region, registry_id)

    res = client.bind_device_to_gateway(
        request={"parent": parent, "gateway_id": gateway_id, "device_id": device_id}
    )

    print(f"Device Bound in IoT {res}")


def _get_iot_gateway_configuration(
        gateway_id: str,
        registry_id: str,
        project_id: str,
        cloud_region: str,
) -> Dict:
    """Returns the Gateway configuration in a dict"""
    iot_client = iot_v1.DeviceManagerClient()
    device_path = iot_client.device_path(project_id, cloud_region, registry_id, gateway_id)

    gateway_device: iot_v1.Device = iot_client.get_device(request={"name": device_path})
    current_config: iot_v1.DeviceConfig = gateway_device.config
    config_json: dict = json.loads(
        current_config.binary_data.decode()
    )
    return config_json


def _update_gateway_configuration(
    gateway_id: str,
    registry_id: str,
    project_id: str,
    cloud_region: str,
    configuration: Dict
) -> None:
    iot_client = iot_v1.DeviceManagerClient()
    device_path = iot_client.device_path(
        project_id,
        cloud_region,
        registry_id,
        gateway_id
    )
    data = json.dumps(
        configuration,
        indent=4
    ).encode()

    iot_client.modify_cloud_to_device_config(
        request={
            "name": device_path,
            "binary_data": data,
            "version_to_update": 0
            # https://googleapis.dev/python/cloudiot/latest/iot_v1/types.html#google.cloud.iot_v1.types.ModifyCloudToDeviceConfigRequest.version_to_update
        }
    )

    print(f"Successfully updated Cloud IoT config for Gateway: {gateway_id}")


def update_paired_devices(event, context):
    """This function is the function body deployed to run in the cloud functions
    runtime.

    A detailed page with all the context can be found here:
        https://cloud.google.com/functions/docs/calling/pubsub
    """
    if "data" in event:
        # ================== Parses event received data====================================
        state_data = base64.b64decode(event["data"]).decode()
        attributes = event.get("attributes", None)
        if not attributes:
            print("Unexpected error: No attributes received in event.", file=sys.stderr)
            exit(1)

        json_data = json.loads(state_data, strict=False)
        print(f"Received JSON data: {json_data}")

        payload: PairedEdgesPayload = PairedEdgesPayload.build_from_dict(json_data)
        if not payload:
            print(
                "Unexpected error: could not parse PairedEdgesPayload in event.",
                file=sys.stderr,
            )
            exit(1)

        # ================== Pulls configuration documents from DB====================================
        # In this point, we'll pull from Firestore the existing EdgeDescriptors.
        # Non-virtual Edges will be there, those need to be updated.
        # Virtual Edges will not be found, those need to be created.
        db = firestore.Client()

        current_configuration = pull_current_configuration_from_firestore(db, payload.tenant_id)
        if not current_configuration:
            print(
                "Unexpected error: No current configuration found in database",
                file=sys.stderr,
            )
            exit(1)

        edges_ids = list(
            map(lambda edge_descriptor: edge_descriptor.logical_id, payload.devices)
        )

        edges_documents: List[firestore.DocumentSnapshot] = _get_edges_from_db(
            db, current_configuration.reference, edges_ids
        )

        if not edges_documents:
            print(
                "Unexpected error: No edges found in configuration for the received edges logical_ids."
            )
            exit(1)

        # ================== Builds associative tables====================================
        edges_documents_table: Dict = dict(
            map(
                lambda edge_doc: (edge_doc.to_dict()["logical_id"], edge_doc),
                edges_documents,
            )
        )

        # ================== Update Firestore documents ==================================
        not_found_edges: List[EdgeDescriptor] = []
        actuator_edge_descriptor_doc: firestore.DocumentSnapshot = None
        for edge_descriptor in payload.devices:
            if edge_descriptor.logical_id in edges_documents_table:
                document = edges_documents_table[edge_descriptor.logical_id]
                parameters_dump = json.dumps(edge_descriptor.edge_parameters, cls=ComplexEncoder)
                document.reference.update({"parameters": json.loads(parameters_dump)})
                print(
                    f"Updated Edge in database: {edge_descriptor.logical_id}. Document ID: {document.id}"
                )
                if edge_descriptor.is_actuator_device():
                    actuator_edge_descriptor_doc = document
            else:
                not_found_edges.append(edge_descriptor)

        print("All existing Edges were updated")

        if not_found_edges:
            print(f"Not found virtual edges will be created")

            if not actuator_edge_descriptor_doc:
                print("Virtual Edges received, but the parent EdgeDescriptor for the actuator"
                      "was not found.", file=sys.stderr)
                db.close()
                exit(1)

            else:
                # TODO: Handle Errors with Firestore.
                created_edges = _create_new_virtual_edges(db, current_configuration.reference,
                                          actuator_edge_descriptor_doc.get("is_of_kind_ref"),
                                          not_found_edges)

                _add_edges_to_machines(db, current_configuration.reference, payload.machines, created_edges)

        db.close()

        # ================= Gateway IoT Device Configuration =============================
        gateway_id: str = attributes['deviceNumId']
        registry_id: str = attributes['deviceRegistryId']
        cloud_region: str = attributes['deviceRegistryLocation']
        project_id: str = attributes['projectId']

        gateway_conf_dict: Dict = _get_iot_gateway_configuration(gateway_id,
                                                                 registry_id,
                                                                 project_id,
                                                                 cloud_region)

        edges_descriptors_json = list(map(lambda edge_desc: json.dumps(
            edge_desc, cls=ComplexEncoder, indent=4
        ), payload.devices))

        edges_descriptors_json = list(map(lambda edge_des_dumped: json.loads(
            edge_des_dumped
        ), edges_descriptors_json))

        gateway_conf_dict.update({
            'edges': edges_descriptors_json
        })

        machines_metadata_jsons = list(map(
            lambda machine_metadata: json.dumps(
                machine_metadata, cls=ComplexEncoder, indent=4
            ), payload.machines
        ))

        machines_metadata_jsons = list(map(
            lambda machine_dumped: json.loads(machine_dumped),
            machines_metadata_jsons
        ))

        gateway_conf_dict["machines"] = machines_metadata_jsons

        # Bind and create new devices
        for edge_descriptor in not_found_edges:
            _bind_device_to_gateway(project_id, cloud_region, registry_id,
                                    edge_descriptor.logical_id, gateway_id)

        _update_gateway_configuration(
            gateway_id, registry_id,
            project_id, cloud_region, gateway_conf_dict
        )

        db.close()
    else:
        print("Unexpected error: No data received in event.", file=sys.stderr)



"""
This module exports the GCP Cloud function to update the Cloud IoT Core Gateway devices
based on the Firestore project configuration, specifically the most recent one.
"""

# =========================== Imports ===============================
from txp.cloud import settings
from txp.common.json_complex_encoder import ComplexEncoder
from txp.common.configuration import GatewayConfig
from txp.common.configuration import JobConfig
from txp.cloud import (
    pull_docs_from_collection_associated_with_configuration
)
from txp.common.edge import EdgeDescriptor, MachineMetadata
import google.cloud.firestore as firestore
from typing import Dict, List
from google.cloud import iot_v1
import json
import logging

log = logging.getLogger(__name__)

# =========================== Constant Data ===============================
manual_configuration_collection = (
    settings.firestore.manual_configuration_edit_collection
)
config_form_document = settings.firestore.config_form_document
machines_group_collection_name = settings.firestore.machines_group_collection
machines_collection_name = settings.firestore.machines_collection
gateways_collection_name = settings.firestore.gateways_collection
devices_collection_name = settings.firestore.devices_collection
edges_collection_name = settings.firestore.edges_collection
jobs_collections_name = settings.firestore.jobs_collection
configurations_collection_name = settings.firestore.configurations_collection
zoned_datetime_str_format = settings.firestore.datetime_zoned_format

configuration_reference_field_path = "configuration_ref"


def _get_edges_descriptor(
        devices_docs: List[firestore.DocumentSnapshot], edges_docs: List[firestore.DocumentSnapshot]
) -> List[EdgeDescriptor]:
    """Utilitary function to build the EdgeDesctiptor's instance for the Gateway IoT Configuration"""
    device_dict: Dict[str, firestore.DocumentReference] = {}
    for device in devices_docs:
        device_dict[device.to_dict()["kind"]] = device.to_dict()

    edges_descriptors: List[EdgeDescriptor] = []
    for edge in edges_docs:
        try:
            device_kind = edge.to_dict()["device_kind"]
            device = device_dict[device_kind]
            if "type" in edge.to_dict():
                device_type = edge.get("type")
            elif "device_type" in edge.to_dict():
                device_type = edge.get("device_type")
            else:
                device_type = device["type"]

            edge_parameters: Dict = {**device["parameters"], **edge.to_dict()["parameters"]}
            edge_perceptions: Dict = device["perceptions"]
            edge_descriptor = EdgeDescriptor(
                edge.to_dict()["logical_id"], device_kind, device_type, edge_parameters, edge_perceptions
            )
            edges_descriptors.append(edge_descriptor)
            print(f"Successfully built the Edge {edge.id}: {edge_descriptor.logical_id} with type: {edge_descriptor.device_type}")
        except:
            edges_descriptors.append(None)

    return edges_descriptors


def _update_gateway_configuration(
        iot: iot_v1.DeviceManagerClient,
        configuration_id: int,
        job: JobConfig,
        gateway_config: GatewayConfig,
        machines: List[MachineMetadata],
        edges: List[EdgeDescriptor]
) -> None:
    """Updates the configuration of the Gateway in the IoT Core service."""
    device_path = iot.device_path(
        gateway_config.project_id,
        gateway_config.cloud_region,
        gateway_config.registry_id,
        gateway_config.gateway_id
    )

    data = json.dumps(
        {
            "configuration_id": configuration_id,
            "job": job,
            "gateway": gateway_config,
            "machines": machines,
            "edges": edges
        },
        cls=ComplexEncoder, indent=4
    ).encode()

    data

    iot.modify_cloud_to_device_config(
        request={
            "name": device_path,
            "binary_data": data,
            "version_to_update": 0
            # https://googleapis.dev/python/cloudiot/latest/iot_v1/types.html#google.cloud.iot_v1.types.ModifyCloudToDeviceConfigRequest.version_to_update
        }
    )


def _extract_edges_ids_from_machine(machine_doc: firestore.DocumentSnapshot):
    edges_ids: List[str] = []
    for edge_doc_ref in machine_doc.get("associated_with_edges"):
        edges_ids.append(edge_doc_ref.get().get('logical_id'))
    return edges_ids


# ======================== Main Function Body to export ============================================
def update_iot_gateways_conf(data, context):
    """The cloud function body defined in this module, to be deployed as a GCP cloud function.

    Documentation reference for the received arguments as result of a Firestore trigger:
        https://cloud.google.com/functions/docs/calling/cloud-firestore#event_structure
    """
    db = firestore.Client()

    log.info("Connection to Firestore established.")

    gateway_document_name = data["value"]["name"]  # This keys are guaranteed by Firestore trigger payload
    gateway_document: firestore.DocumentSnapshot = db.document(gateway_document_name).get()
    job_document: firestore.DocumentSnapshot = gateway_document.get("has_job").get()

    machines: List[firestore.DocumentSnapshot] = list(map(
        lambda machine_doc_ref: machine_doc_ref.get(),
        gateway_document.get('machines')
    ))

    edges: List[firestore.DocumentSnapshot] = []

    for machine in machines:
        edges = edges + (list(map(
            lambda edge_doc_ref: edge_doc_ref.get(),
            machine.get("associated_with_edges")
        )))

    edges = list(set(edges))

    configuration = gateway_document.get('configuration_ref')

    if not configuration:
        log.error(f"No Configuration reference document was found in "
                  f"the Gateway {gateway_document.get('gateway_id')}")
        return 1

    tenant_doc_ref = configuration.get().get('tenant_ref')
    tenant_id = tenant_doc_ref.get().get('tenant_id')

    log.info(f'Tenant ID found for Gateway update: {tenant_id}')

    devices_documents = pull_docs_from_collection_associated_with_configuration(
        db, devices_collection_name, configuration
    )

    # Build TXP objects instances.
    configuration_id = configuration.get().to_dict().get('configuration_id', None)
    if not configuration_id:
        log.error("Configuration obtained from firestore does not have the expected configuration_id field.")
        return 1

    configuration_id = configuration_id

    machines_ids: List[str] = list(map(
        lambda machine_doc: machine_doc.to_dict()['machine_id'], machines
    ))

    gateway_config: GatewayConfig = GatewayConfig.build_from_dict(
        {**gateway_document.to_dict(), **{"machines_ids": machines_ids}}
    )

    job_config: JobConfig = JobConfig.build_from_dict(job_document.to_dict())
    if not job_config:
        log.error("The JOB obtained from Firestore could not be parsed")
        return 1

    if gateway_config is None:
        log.error("Gateway configurations obtained from Firestore could not be parsed")
        return 1

    machines_metadata: List[MachineMetadata] = list(map(
        lambda machine_document: MachineMetadata.build_from_dict(
            {**machine_document.to_dict(), **{"edges_ids": _extract_edges_ids_from_machine(machine_document)}}
        ),
        machines
    ))

    for machine in machines_metadata:
        if machine is None:
            log.error("MachinesMetadata configurations obtained from Firestore could not be parsed")
            return 1

    edges_descriptors: List[EdgeDescriptor] = _get_edges_descriptor(devices_documents, edges)

    for edge_descriptor in edges_descriptors:
        if edge_descriptor is None:
            log.error("EdgeDescriptor configurations obtained from Firestore could not be parsed")
            return 1

    iot = iot_v1.DeviceManagerClient()

    _update_gateway_configuration(
        iot,
        configuration_id,
        job_config,
        gateway_config,
        machines_metadata,
        edges_descriptors
    )

    print(f"Successfully updated the Gateway: {gateway_config.gateway_id} in IoT")


# ======================== Main to debug locally ============================================
if __name__ == '__main__':
    # debug using the GAO standalone-arm gateway
    update_iot_gateways_conf(
        {
            'value': {
                'name': 'gateways/6QQ6anowueUC25seUfnS'
            }
        },
        {})



"""
This module defines a common Camara driver this class is designed to wrap up
the camera functionalities to be used as a resource in a txp driver class.
this class does not innherits from driver.py which is driver txp class.
"""
# ============================ imports =========================================
import logging
from enum import Enum
from typing import Optional
try:
    import pyudev
except ModuleNotFoundError:
    pass
import cv2
from txp import settings

log = logging.getLogger(__name__)
log.setLevel(settings.txp.general_log_level)


class CameraType(Enum):
    """CameraType.
    All possible camera types will be listed here.
    NORMAL: regular camera
    THERMOGRAPHIC: thermal camera
    TODO: this intends to  have in the enumerated type the enum including the model

    """
    def __new__(cls, *values):
        """This defines the cannonical value (int) for the enum, and allows for other
        values. Useful for human readable strings.

        Reference for this solution:
            https://stackoverflow.com/questions/43202777/get-enum-name-from-multiple-values-python
        """
        obj = object.__new__(cls)
        obj._value_ = values[0]
        for other_value in values[1:]:
            cls._value2member_map_[other_value] = obj
        obj._all_values = values
        # human readable value useful for sending payload information to other systems
        obj.human_readable_value = values[1]
        return obj

    @classmethod
    def has_value(cls, value):
        return value in cls._value2member_map_

    def __str__(self):
        return self.human_readable_value

    NORMAL = 0, 'Normal', 'normal'
    THERMOGRAPHIC = 1, 'Thermographic', 'thermographic'
    THERMOCOUPLE = 2, 'Thermocouple', 'thermocouple'


class CameraHandlerBase:

    def __init__(self, camera_type: str, camera_id: str):
        if not CameraType.has_value(camera_type):
            log.error(f"Camera Handler does not recognize unknown camera type: {camera_type}")
            raise ValueError(f"Unknown CameraType: {camera_type}")

        self._camera_handler = None
        self._camera_id = camera_id
        self._camera_type = CameraType(camera_type)

    @classmethod
    def detect_usb_camera(cls) -> Optional[str]:
        """Detects the connection of a USB Camera connected to some of the USB ports.

        Returns:
            The camera_id detected. None if nothing was detected.
        """
        result = list()
        context = pyudev.Context()
        for device in context.list_devices(
            subsystem="video4linux", ID_V4L_CAPABILITIES=":capture:"
        ):
            links = device.get("DEVLINKS").split()
            physical_link = list(filter(lambda link : "/by-id/" in link,links))
            link_name = physical_link.pop()
            result.append(link_name)
            #result.append(device.get("DEVLINKS"))
            log.info(
                f"CameraHandler: USB Camera {device.get('ID_VENDOR_ID', None)}:{device.get('ID_MODEL_ID', None)} "
                f"found from {device.get('ID_VENDOR', None)}"
                f"Device name {link_name}"
            )


        return result[-1] #last item mounted

    def connect(self):
        """Connects to the USB camera with the corresponding physical id.

        Returns:
            True if camara was correctly detected and connected
        """
        log.debug(f"Connecting USB camera")
        if not self._camera_handler:
            self._camera_handler = cv2.VideoCapture(self._camera_id)

        if not self._camera_handler.isOpened():
            log.error(f"Error connecting to the USB physical_id: {self._camera_id}:")
            return False

        return True

    def disconnect(self):
        """
            Disconnects the USB camera and release the handler.
        """
        if self._camera_handler:
            self._camera_handler.release()
            log.info(f"USB camera {self._camera_id} was disconnected")

    def get_camera_id(self):
        """
        :return: returns the camera camera_id
        """
        return self._camera_id

    def get_camera_type(self):
        """

        :return: camera type
        """
        return self._camera_type

    def is_connected(self):
        """

        :return: True if resource is connected
        """
        if self._camera_handler is not None:
            camera_status = self._camera_handler.isOpened()
            log.debug("camera status {}".format(camera_status))
            return camera_status
        return False

    def get_image(self):
        for i in range(1,20):
            ret, image = self._camera_handler.read()
        log.info(f"Image taken from {self._camera_id}")
        if not ret:
            log.error("Error reading image")
            return []  # Calling cv2.imencode with empty image will crash. Return here.

        ret, image_jpeg = cv2.imencode('.jpeg', image)
        if not ret:
            log.error("Error encoding JPEG image")

        log.info("returning JPEG")
        return list(image_jpeg)




"""
This module contains classes to support the Driver conceptualization.
"""

# ============================ imports =========================================
import copy
from enum import Enum
from dataclasses import dataclass
from abc import ABC, abstractmethod
from txp.common.edge import EdgeDescriptor
from txp.common.edge import DeviceId
from txp.common.edge import Signal
from txp.cloud.common.utils import SignalMode
from txp.devices.package import (
    GatewayPackage,
    GatewayPackageMetadata,
)
from txp.devices.package_queue import PackageQueueProxy
from txp.devices.sampling_task import SamplingTask
import numpy as np
import datetime
import queue
import threading
import time
from typing import List, Dict
from txp.common.config import settings
import logging
from collections import deque
import pytz

log = logging.getLogger(__name__)
log.setLevel(settings.txp.general_log_level)


# ==============================================================================
# Class to encapsulate the DriverState
# ==============================================================================
class DriverState(Enum):
    """A Driver state encapsulation to propagate state across components.
    STATES ARE:
    DISCONNECTED: means that the driver is disconnected
    CONNECTED: driver is connected and ready to sample. When driver is sampling it goes from SAMPLING to CONNECTED
               during the sampling window. Meaning that when  connected there are no pending tasks to be performed
               by the driver
    SAMPLING: means that the driver is currently collecting signals measures
    RECOVERING: this can be used by the host to perform an automatic recovery action from the host and avoid to
                notify that the driver is disconnected if it can be reconnected automatically.
    ERRORED: meaning that it wasnt possible to determine a knonw state for the driver. like CONNECTED or DISCONNECTED.
            this is probably the persisten state that must trigger a human interaction with the physical driver
            to determine the source of the error.
    Note: This class is expected to grow with time.

    """
    DISCONNECTED = 0, 'Disconnected'
    CONNECTED = 1, 'Connected'
    SAMPLING = 4, 'Sampling'
    RECOVERING = 2, 'Recovering'
    ERRORED = 3, 'Errored'  # this state means that the device was connected and not having data on a sampling window

    def __new__(cls, *values):
        """This defines the cannonical value (int) for the enum, and allows for other
        values. Useful for human readable strings.

        Reference for this solution:
            https://stackoverflow.com/questions/43202777/get-enum-name-from-multiple-values-python
        """
        obj = object.__new__(cls)
        obj._value_ = values[0]
        for other_value in values[1:]:
            cls._value2member_map_[other_value] = obj
        obj._all_values = values
        # human readable value useful for sending payload information to other systems
        obj.human_readable_value = values[1]
        return obj


@dataclass
class StateReport:
    """
    this class has 2 items of information.
    serial: logical serial id of the driver
    state: ReportingDriverState
    #TODO: Include Task ID for imcompleted task
    """
    logical_id: str
    driver_state: DriverState
    since_datetime: datetime.datetime = None

    def __post_init__(self):
        # Compute datetime for the state
        self.since_datetime = datetime.datetime.now(
            pytz.timezone(settings.gateway.timezone)
        ).replace(microsecond=0)


# ==============================================================================
# Definition of the Base Driver class
# ==============================================================================
class Driver(ABC):
    """This Driver is the generic concept of a software component that interacts and
    represents one edge device.

    This Driver base class takes care of communication with the associated Gateway. Doing so,
    child drivers will only need to deal with interactions related to the specific hardware device,
    while delegating information sending to the base implemented methods.
    This Driver base supports an internal state machine, which is used to provide
    built-in states and transitions for the derived classes.
    """

    # each package size should be within [170000, 180000]
    _max_size_packages = (170000, 180000)

    # Standard perception dimensions fields common to all edges
    PERCEPTION_DIMENSIONS_FIELD = "dimensions"
    PERCEPTION_SAMPLING_FREQUENCY_FIELD = "sampling_frequency"

    # All drivers must define a physical_id edge parameter.
    _PHYSICAL_ID_PARAM_FIELD = "physical_id"

    def __init__(
            self,
            edge_descriptor: EdgeDescriptor,
            package_queue: PackageQueueProxy,
            on_driver_state_changed: callable,
            completed_clocks_notif_queue: queue.Queue
    ):
        """The Driver initialization process.

        After setting up the attributes for the driver, the machine state is declared.
        However, the Driver instance should call the `connect_device()ß` method to
        initialize the state in the derived class. This allows the derived driver to
        setup any required code at initialization time before trying to connect.

        Args
            edge_descriptor: The description of the edge handled by this driver.
            sampling_window: The sampling configuration window set in the gateway
                for this driver.
            package_queue: The Package queue to send produced packages by this Driver.
            edge_descriptor: The EdgeDescriptor for the edges connected to this driver
            host_notification_queue: The host queue where this driver will publish the
                appropriated reports everytime it finishes an observation window and
                produces a GatewayPackage.
        """
        # Data
        self.logical_id: str = edge_descriptor.logical_id

        self._physical_id: str = edge_descriptor.edge_parameters[self._PHYSICAL_ID_PARAM_FIELD]
        """_physical_id(str): The string representation of the physical ID of the edge, used to define 
        the complete DeviceId"""

        self.device_kind: str = edge_descriptor.device_kind
        self.edge_parameters: Dict = edge_descriptor.edge_parameters

        self.perceptions: Dict = edge_descriptor.perceptions
        """perceptions: The perception dimensions supported by the driver. They are received as a field inside the 
        edge_parameters constructor parameter"""

        if not self.perceptions:
            log.error(
                f"Driver: Empty perceptions received in Edge Descriptor."
            )
            raise EdgeParameterNotFoundError(
                f"Driver: Empty perceptions received in Edge Descriptor"
            )

        self.package_queue: PackageQueueProxy = package_queue
        self.pending_tasks_queue: queue.Queue = queue.Queue()
        self.driver_state = DriverState.DISCONNECTED
        self.on_driver_state_changed = on_driver_state_changed
        self._update_state_lock = threading.Lock()

        """
        Window sampling control attributes
        """
        self._sampling_thread: threading.Thread = None
        """_sampling_thread: is the thread of control that will be used to 
        monitor the sampling process according to the defined cadence."""

        self._sampling_thread_pill: threading.Event = None
        """_sampling_thread_pill: A Kill pill for the thread"""

        self._observation_pulse_signal: threading.Event = threading.Event()
        """_observation_pulse_signal: This control event is used by the upper layers to 
        authorize the driver to start a observation pulse at some point in time."""

        self._stop_current_pulse_signal: threading.Event = threading.Event()
        """_stop_current_pulse_signal: This attribute allows the driver to know if somebody stopped its current
        sampling clock performed by the method _start_sampling_collect."""

        self._completed_clocks_notif_queue: queue.Queue = completed_clocks_notif_queue

        self._sampling_window_finished: threading.Event = threading.Event()
        """_sampling_window_finished: an event that allows the driver to know when a sampling 
        window was completed and the it can proceed to send the packages."""

    @abstractmethod
    def observation_pulse(self) -> int:
        """Returns the Observation pulse of the Driver which is used for the
        synchronization tasks clock in the system

        Returns:
            The observation pulse in seconds.
        """
        pass

    @classmethod
    @abstractmethod
    def device_name(cls) -> str:
        pass

    @classmethod
    @abstractmethod
    def is_virtual(cls) -> bool:
        """Each Driver should declare in the implementation if it is a
        virtual driver or not.
        There's a correspondence between the EdgeDescriptor device_kind and
        the Driver's device_name() method. By knowing the device_kind of an EdgeDescriptor,
        is possible to know if the Driver is virtual or not.
        """
        pass

    def enable_observation_pulse(self) -> None:
        log.info(f"Driver {self.logical_id} was enabled for a new driver's observation period")
        self._observation_pulse_signal.set()

    def disable_current_pulse(self) -> None:
        log.info(f"Driver {self.logical_id} was requested to stop its current sampling collect pulse.")
        self._stop_current_pulse_signal.set()

    def get_complete_device_id(self) -> DeviceId:
        """Returns the defined DeviceId for the edge device.

        The DeviceId for an edge is the combination of the logical_id and the physical_id.
        This association is made through the devices pairing process.
        """
        return DeviceId(self.logical_id, self._physical_id)

    def get_descriptor(self) -> EdgeDescriptor:
        return EdgeDescriptor(
            self.logical_id,
            self.device_kind,
            self.device_name(),
            self.edge_parameters,
            self.perceptions
        )

    def get_state(self) -> DriverState:
        return self.driver_state

    def update_driver_state(self, driver_state: DriverState):
        try:
            self._update_state_lock.acquire(True)
            if self.driver_state == DriverState.DISCONNECTED:
                if driver_state != DriverState.CONNECTED:
                    log.error("{} Invalid transition current state is {} and only allows transitions to: {}".format(
                        self.logical_id, self.driver_state, DriverState.CONNECTED.name))
        finally:
            self.driver_state = driver_state
            log.info("{} device state is {}".format(self.logical_id, self.driver_state))
            self._update_state_lock.release()

    def _send_package_to_gateway(self, package: GatewayPackage):
        """Send the produced package to the central PackageQueue."""
        self.package_queue.add_package(package)

    def receive_sampling_task(self, task: SamplingTask):
        """Receives a new sampling task for this Driver.
        The driver implementation ensures to eventually process this task.
        """
        log.debug(f"Driver {self.logical_id} Receive Sampling Task")
        self.pending_tasks_queue.put(task)

    def has_pending_tasks(self) -> bool:
        """Checks if there are pending tasks to be processed.
        :return: True if there are pending tasks in the queue; otherwise False.
        """
        return not self.pending_tasks_queue.empty()

    @abstractmethod
    def connect_device(self):
        """Connect the device handled by the Driver.

        After the connection succeed or fails, the state of the
        driver should change accordingly. The suggested way is
        to use the defined transitions:
            Connection success:
                self._device_connected()
            Connection Fails:
                self._device_disconnected()
        """
        pass

    @abstractmethod
    def disconnect_device(self):
        """Disconnects the device handled by the  Driver.

        After the device gets disconnected, the state of the driver
        should reflect the change. The suggested way is to use the
        defined transition:
            self._device_disconnected()
        """
        pass

    @abstractmethod
    def _start_sampling_collect(self) -> List[Signal]:
        """This method will collect the samples obtained from the edge device and will
        produce the system Signals to send the package to the gateway.

        This method will collect the samples from a driver clock pulse.
        """
        pass

    @abstractmethod
    def _stop_sampling_collect(self):
        """This method will stop the sampling collection in the driver implementation.
        It can be used to clean resources using during the samples collection process.
        """
        pass

    def start_sampling_windows(self):
        """Call this method to start the _sampling_thread"""
        if self.driver_state == DriverState.CONNECTED:
            if self._sampling_thread is None:
                # Clear timer pill from any previous use
                self._sampling_thread_pill = threading.Event()

                self._sampling_thread = threading.Thread(
                    target=self._sampling_windows_loop,
                    name=f"window_timer_thread_{self.logical_id}",
                    daemon=True,
                    args=(self._sampling_thread_pill,),
                )

                log.info(f"Starting a new sampling thread in Driver: {self.logical_id}")
                self._sampling_thread.start()
            else:
                log.warning(
                    "Trying to start a sampling thread. But there is one already started."
                    "The driver must explicitly stop with _stop_sampling() call."
                )
        else:
            log.warning(
                "Cannot start sampling window. Invalid DriverState. Driver serial {} state {}".format(self.logical_id,
                                                                                                      self.driver_state))

    def notify_sampling_window_completed(self) -> None:
        log.info(f"Driver {self.logical_id} was notified of sampling window completion")
        self._sampling_window_finished.set()

    def stop_sampling_windows(self):
        """Set the _sampling_thread_pill in order to finish the current
        _sampling_thread"""
        if self._sampling_thread is not None:
            self._sampling_thread_pill.set()

            # Todo: Implement a timeout to return and log a Warning
            while (
                    self._sampling_thread is not None and self._sampling_thread.is_alive()
            ):
                time.sleep(0.1)

            self._sampling_thread = (
                None  # The thread eventually will finish thanks to the pill
            )
            log.info(f"Stopping sampling thread in Driver: {self.logical_id}")
            self._report_connected_state_from_sampling()
        else:
            log.debug("Trying to stop the sampling_thread, but non thread was found.")

    def _report_connected_state_from_sampling(self):
        """
        If previous state is not SAMPLING state change will not be reported
        :return:
        """
        if self.driver_state == DriverState.SAMPLING:
            self._notify_new_driver_state_to_host(DriverState.CONNECTED)  # only go back to connected if it makes sense

    def _sampling_windows_loop(self, stop_pill: threading.Event):
        """The loop executed in the _sampling_thread.

        Here is where samples collection will occur, according to the sample
        cadence defined in the SamplingWindow of the Driver and the concrete
        implementation of the methods _start_sampling_collect() and _stop_sampling_collect().
        """
        requested_to_stop: bool = False
        signals = []
        sampling_task: SamplingTask = None

        while True:
            if requested_to_stop:
                break

            self._change_to_sampling_state()
            # No sampling task to collect.

            # Wait for the upper levels to authorize the next observation execution.
            while not self._observation_pulse_signal.is_set():
                time.sleep(0.05)  # TODO: this time should be configurable
                if self._sampling_window_finished.is_set():
                    requested_to_stop = True
                    break

                if not stop_pill.is_set():
                    continue

                else:
                    requested_to_stop = True
                    break

            else:
                self._observation_pulse_signal.clear()

                sampling_task: SamplingTask = self.pending_tasks_queue.get()

                log.info(f"Driver {self.logical_id} will perform a driver observation clock pulse")

                clock_pulse_time_finished = threading.Event()
                dead_time_timer = threading.Timer(
                    self.observation_pulse(),
                    lambda e: e.set(),
                    (clock_pulse_time_finished,),
                )
                dead_time_timer.start()

                # TODO: The driver start_sampling_collect should contemplate an interruption if somebody requested
                #   to stop. Concrete implementations can handle that.
                signals_in_pulse = self._start_sampling_collect()
                signals = signals + signals_in_pulse

                # Extra check to ensure that sampling time has finished.
                while not clock_pulse_time_finished.is_set():
                    if stop_pill.is_set() or self._sampling_window_finished.is_set():
                        requested_to_stop = True
                        log.warning(f"Driver {self.logical_id} requested to stop its sampling while in-middle "
                                    f"of sampling window.")
                        break

                self._stop_sampling_collect()

                log.info("{} Signals measured in last clock pulse: {}.".format(
                    self.logical_id, len(signals_in_pulse)))

                log.debug("{} driver finished a clock pulse at {} hours".format(
                    self.logical_id, datetime.datetime.now().time()))

                self._completed_clocks_notif_queue.put(
                    self.logical_id)  # Puts the logical id when a clock tick finishes.

        if not sampling_task:
            log.warning(f"Driver {self.logical_id} No collected sampling task from Sampling tasks queue."
                        f" No packages to send.")
        else:
            gateway_packs: List[GatewayPackage] = self._build_gateway_packages(signals, sampling_task)
            if len(gateway_packs) > 0:
                log.info("Driver collected {} signals and generated {} subpackages".format(
                    len(signals), len(gateway_packs)
                ))
                for pack in gateway_packs:
                    log.info(f"Package produced by driver {self.logical_id}: "
                             f"observation_timestamp: {pack.metadata.sampling_window.observation_timestamp} - "
                             f"sampling_window_index: {pack.metadata.sampling_window.sampling_window_index} -"
                             f"num_sampling_windows: {pack.metadata.sampling_window.number_of_sampling_windows}")
                    self._send_package_to_gateway(pack)
            else:
                log.info("{} driver collected 0 samplings in the sampling window. Will try to reconnect".format(
                    self.logical_id))
                self._process_no_signals_in_window_error()

        self._observation_pulse_signal.clear()
        self._stop_current_pulse_signal.clear()
        self._sampling_window_finished.clear()
        self._report_connected_state_from_sampling()

    def _change_to_sampling_state(self):
        if self.driver_state == DriverState.CONNECTED:
            self._notify_new_driver_state_to_host(DriverState.SAMPLING)

    def _process_no_signals_in_window_error(self):
        """This method implements the no signals case (driver is not sending any signals).
        """
        if self.driver_state != DriverState.CONNECTED and self.driver_state != DriverState.SAMPLING:
            log.warning("Driver is already disconnected cannot try to recover")
        else:
            self._notify_new_driver_state_to_host(DriverState.RECOVERING)

    def _notify_new_driver_state_to_host(self, driver_state):
        self.on_driver_state_changed(self.logical_id, driver_state)

    @staticmethod
    def _get_bytes_in_signal(signal):
        """ Calculates the size in bytes of a signal
            Args:
                signal: sampled signal
        """
        bytes_in_signal = 0
        if signal.is_scalar():
            bytes_in_signal = 4
        else:
            for sample in signal.samples:
                bytes_in_signal += len(sample) * sample.itemsize
        return bytes_in_signal

    def _save_package(self, gateway_packages, sampling_task, signals_to_pack, current_package_index,
                      package_id_prefix, previous_part_package_index):
        """ Saves package into gateway_packages list
            Args:
                gateway_packages: list containing all list of packages
                sampling_task: current task
                signals_to_pack: signals that will go into the package
                current_package_index: current index into list of packages,
                previous_part_package_index: previous index in the list of packages
        """
        gateway_packages.append(GatewayPackage(
            sampling_task.configuration_id,
            signals_to_pack.copy(),
            GatewayPackageMetadata(
                self.get_descriptor(), sampling_task.sampling_window,
                package_id_prefix + str(current_package_index),
                previous_part_package_index, sampling_task.tenant_id
            ),
        ))
        signals_to_pack.clear()

    @staticmethod
    def _join_signals_of_sampling_window(sampled_signals):
        """This function joins VECTOR signals of the same perception of a sampling window into a single big signal
            If the signal is not a VectorSignal type, it's samples are not joined together.

            Args:
                sampled_signals: all signals sampled in a sampling windows interval
        """
        nonvector_signals = [signal for signal in sampled_signals if SignalMode.is_time_only(signal.mode().value)]
        groups = {}
        for signal in sampled_signals:
            if SignalMode.is_time_only(signal.mode().value):
                continue
            if signal.perception_name() in groups:
                for i, dimension in enumerate(signal.samples):
                    groups[signal.perception_name()].samples[i] = np.concatenate(
                        (groups[signal.perception_name()].samples[i], dimension), axis=None)
            else:
                groups[signal.perception_name()] = signal
        return deque(list(groups.values()) + nonvector_signals)

    def _split_signal(self, signal, accumulated_bytes):
        """Splits a signal according to maximum size interval
            The length of a package should be withing the interval:
                [_max_size_packages[0], _max_size_packages[1]]

                since accumulated_bytes >= _max_size_packages[0] the signal should be split at size:
                    (_max_size_packages[1] - accumulated_bytes) / (signal.perception_dimensions[0] *
                                                                                            signal.samples.itemsize))
            Args:
                signal: Big signal whose size cannot fit into desired interval
                accumulated_bytes: already accumulated bytes into signals_to_pack
        """
        upper_bound = self._max_size_packages[1] - accumulated_bytes
        idx = int(upper_bound / (signal.perception_dimensions[0] * signal.samples[0].itemsize))
        rest = copy.deepcopy(signal)
        for i, dimension in enumerate(signal.samples):
            rest.samples[i] = np.copy(dimension[idx:])
            signal.samples[i] = dimension[:idx]
        return rest

    def _build_gateway_packages(self, sampled_signals, sampling_task: SamplingTask):
        """Build a set List of Gateway Packages based on the received sampled signals by the driver.
            The list of all signals will be distributed in a list of Gateway Packages. The desired behavior is:
                - Put all the possible information inside the Same Gateway Packages. Signals can be spread out
                    in different packages if required.
                - If a signal is spread out in different packages, those packages are referenced as a single linked list
                    and the package Timestamp will be shared for all those packages.
                - If all the signals in the packages are distributed without being shared in different packages,
                    then the packages don't need to be related among them.
            Args:
                sampled_signals: all signals sampled in a sampling windows interval
                sampling_task: sampling task of signals
        """
        sampled_signals = self._join_signals_of_sampling_window(sampled_signals)
        gateway_packs = []
        accumulated_bytes = 0
        signals_to_pack = []
        package_timestamp = 0
        current_package_index = 0
        gateway_package_index = -1

        while len(sampled_signals):
            signal = sampled_signals.popleft()
            sample_bytes = self._get_bytes_in_signal(signal)
            accumulated_bytes += sample_bytes
            if accumulated_bytes >= self._max_size_packages[0] or \
                    (accumulated_bytes <= self._max_size_packages[0] and current_package_index != 0):
                if accumulated_bytes <= self._max_size_packages[1]:
                    signals_to_pack.append(signal)
                    if current_package_index == 0:
                        package_timestamp = time.time_ns()
                    else:
                        gateway_packs[gateway_package_index].metadata.previous_part_index = current_package_index
                    package_id_prefix = f"{package_timestamp}_"
                    previous_part_package_index = 0 if current_package_index == 0 else current_package_index - 1
                    self._save_package(gateway_packs, sampling_task, signals_to_pack, current_package_index,
                                       package_id_prefix, previous_part_package_index)
                    current_package_index = 0
                else:
                    if current_package_index == 0:
                        package_timestamp = time.time_ns()
                        gateway_package_index = len(gateway_packs)
                        if len(signals_to_pack):
                            package_timestamp = time.time_ns()
                            package_id_prefix = f"{package_timestamp}_"
                            self._save_package(gateway_packs, sampling_task, signals_to_pack, 0, package_id_prefix, 0)
                            accumulated_bytes = 0
                            sampled_signals.appendleft(signal)
                            continue

                    accumulated_bytes -= sample_bytes
                    rest = self._split_signal(signal, accumulated_bytes)
                    signals_to_pack.append(signal)
                    package_id_prefix = f"{package_timestamp}_"

                    if not SignalMode.is_time_only(signal.mode().value) and \
                            settings.gateway.streaming_mode == "REALTIME":
                        assert current_package_index == 0
                        self._save_package(gateway_packs, sampling_task, signals_to_pack, current_package_index,
                                           package_id_prefix, 0)
                    else:
                        self._save_package(gateway_packs, sampling_task, signals_to_pack, current_package_index,
                                           package_id_prefix, current_package_index - 1)
                        current_package_index += 1
                        sampled_signals.appendleft(rest)
                accumulated_bytes = 0
            else:
                signals_to_pack.append(signal)

        if len(signals_to_pack) > 0:
            if current_package_index == 0:
                package_timestamp = time.time_ns()
            else:
                gateway_packs[gateway_package_index].metadata.previous_part_index = current_package_index
            package_id_prefix = f"{package_timestamp}_"
            previous_part_package_index = 0 if current_package_index == 0 else current_package_index - 1
            self._save_package(gateway_packs, sampling_task, signals_to_pack, current_package_index,
                               package_id_prefix, previous_part_package_index)

        return gateway_packs


# ==============================================================================
# Custom Exceptions for Driver class
# ==============================================================================


class EdgeParameterNotFoundError(Exception):
    """This exception should be thrown when a Driver does not found an expected
    value in the EdgeDescriptor.edge_parameters dictionary.
    """
    pass


class MalformedEdgeParameterError(Exception):
    """This exception should be thrown when a Driver finds a malformed value in the
    EdgeDescriptor.edge_parameters dictionary.
    """
    pass



from abc import ABC, abstractmethod
from joblib import dump, load
from dataclasses import dataclass


@dataclass
class PolicyStorage:
    handler: any
    trained: bool
    accuracy: float


class Policy(ABC):

    def __init__(self, policy_file=None):

        if policy_file is not None:
            print(f"Restoring policy from {policy_file}")
            self.clf = load(policy_file)
            self.trained = True
        else:
            self.trained = False
            self.clf = None
        self.accuracy = -1

    @abstractmethod
    def train(self, dataset, target, test_size=0.3):
        pass

    @abstractmethod
    def predict(self, tensor):
        pass

    @abstractmethod
    def name(self):
        pass

    def save(self) -> PolicyStorage:
        return PolicyStorage(self.clf, self.trained, self.accuracy)

    def load(self, policy_storage: PolicyStorage):
        print("Restoring policy")
        self.clf = policy_storage.handler
        self.trained = policy_storage.trained
        self.accuracy = policy_storage.accuracy

    def is_trained(self):
        return self.trained



from skimage import segmentation, morphology, color
import numpy as np
import cv2

START_LABEL = 1
BACKGROUND_PIXEL_COLOR = 255  #this value CANNOT be changed otherwise the masking algo wont work.
                              #TODO: test with a withe object.

class SlicPatchImageProcessor:
    def __init__(self,number_of_segmentss):
        self._number_of_segments = number_of_segmentss


    def _calculate_start_end_position(self,block, block_len, pixels_size):
        start = block * block_len
        if (pixels_size - start) < block_len:
            start = pixels_size - block_len
        end = start + block_len
        return start, end

    def _get_b_w_image_from_mask(self,patch_img, masked_p):
        ret = np.ndarray(shape=patch_img.shape,dtype=patch_img.dtype)
        ret[:] = (0,0,0)
        for x in range(patch_img.shape[0]):
            for y in range(patch_img.shape[1]):
                if masked_p[x][y]:
                    ret[x][y] = (255,255,255)
        return  ret

    def _binarize_image(self,masked_p, patch_img):
        b_w_image = self._get_b_w_image_from_mask(patch_img, masked_p)
        gr = cv2.cvtColor(b_w_image, cv2.COLOR_BGR2GRAY)
        _, im = cv2.threshold(gr, 127, 255, cv2.THRESH_BINARY)
        return im

    def _get_slic_unique_values(self,slic_img, patch_img):
        vector = []
        for i in range(1,40):
            mask = slic_img == i
            masked = patch_img[mask]
            if masked.size > 0:
                vector.append(np.mean(masked))
            else:
                vector.append(-1)
        return vector

    def _get_patch_vector(self,img, slic_superpixels_original_image, x_end, x_start, y_block, y_len, y_pixels_size):

        y_start, y_end = self._calculate_start_end_position(y_block, y_len, y_pixels_size)
        slic_img, patch_img, masked_p = self._get_patch_background_foreground(img,
                                                                              (x_start, x_end),
                                                                              (y_start, y_end),
                                                                              slic_superpixels_original_image)
        im = self._binarize_image(masked_p, patch_img)
        moments = cv2.moments(im)
        ordered_moments = dict(sorted(moments.items()))
        huMoments = cv2.HuMoments(moments)
        vector_p = []
        vector_p.extend(list(ordered_moments.values()))
        vector_p.extend(np.concatenate(huMoments).ravel().tolist())
        vector_p.extend(self._get_slic_unique_values(slic_img, patch_img))
        return vector_p

    def _get_dominant_superpixel(self,labeled_slice):
        return np.bincount(labeled_slice.ravel()).argmax()

    def _get_only_foregroung_of_dominant_pixel(self,img_patch, max_pixel, labeled_slice):
        masked_3d = np.broadcast_to( np.atleast_3d(labeled_slice==max_pixel), img_patch.shape)
        return np.where(masked_3d,img_patch,BACKGROUND_PIXEL_COLOR)

    def _get_patch_background_foreground(self, img, x_min_max, y_min_max, slic_superpixels_labels,
                                         start_label=START_LABEL):
        """
        separetares the foreground which is object being looked from the background.

        :param img: numpy array image 2D array pixels representation
        :param x_min_max: (xmin,xmax) tuple of the region in the image in which the
        :param y_min_max:(ymin,ymax) tuple of the region in the image in which the
        :return: labeled slic and the object detected with plain background
        """

        labeled_slice = slic_superpixels_labels[x_min_max[0]:x_min_max[1], y_min_max[0]:y_min_max[1]]

        max_pixel = self._get_dominant_superpixel(labeled_slice)
        img_patch = img[x_min_max[0]:x_min_max[1], y_min_max[0]:y_min_max[1], :]
        sliced_img_f_b = self._get_only_foregroung_of_dominant_pixel(img_patch, max_pixel, labeled_slice)
        masked_patch = self._compute_mask(sliced_img_f_b)

        values = np.unique(masked_patch)

        if len(values) > 1:
            slic_superpixels_labels_patch = segmentation.slic(sliced_img_f_b, n_segments=self._number_of_segments, start_label=start_label,mask=masked_patch)
        else:
            slic_superpixels_labels_patch = segmentation.slic(sliced_img_f_b, n_segments=self._number_of_segments, start_label=start_label)

        return slic_superpixels_labels_patch, sliced_img_f_b,masked_patch

    def _compute_mask(self, img):
        lum = color.rgb2gray(img)
        mask = morphology.remove_small_holes(
            morphology.remove_small_objects(lum < 0.7, 500), 500)
        mask = morphology.opening(mask, morphology.disk(3))
        return mask

    def process_image_patches(self,img,x_len,y_len,blocks_multiplier=1):
        """

        :param img: image to process the metrics based on a patch size
        :param patcth_size: the patch size tuple (Xlen,Ylen)
        :param blocks_multiplier: defines the multiplier of the blocks search based on the min blocks
        :return: an array of patches metrics vector
        """
        vector = []
        mask_general = self._compute_mask(img)
        slic_sp =  segmentation.slic(img, n_segments=self._number_of_segments, start_label=START_LABEL,mask=mask_general)
        x_pixels_size = img.shape[0]
        y_pixels_size = img.shape[1]
        x_blocks_qty = int(x_pixels_size / x_len * blocks_multiplier)
        y_blocks_qty = int(y_pixels_size / y_len * blocks_multiplier)

        for x_block in range(x_blocks_qty):
            x_start,x_end = self._calculate_start_end_position(x_block, x_len, x_pixels_size)
            for y_block in range(y_blocks_qty):
                vector_p = self._get_patch_vector(img,slic_sp, x_end, x_start, y_block, y_len, y_pixels_size)
                vector.extend(vector_p)
        return vector



from txp.web.core_components.main_view import MainView
from txp.web.core_components.app_profiles import AppProfile
import streamlit as st
from txp.cloud.controllers.firestore import *
from txp.common.json_complex_encoder import ComplexEncoder
from typing import Dict
import logging
import json

from txp.common.config import settings

log = logging.getLogger(__name__)
log.setLevel(settings.txp.general_log_level)


class EditConfigurationView(MainView):
    """This View implements use cases to upate configuration of the project model.

    Currently, it only supports the use case to update the Job of a Gateway device.
    """

    def __init__(self):
        super(EditConfigurationView, self).__init__(component_key="edit_configuration")
        self._project_configuration_data: ProjectModel = None
        self._current_gateway: str = ""
        self._text_area_unique_key = f"{id(self)}_job_for_gateway_text_area"
        self._gateway_selectbox_unique_key = f"{id(self)}_gateway_selectbox"
        self._show_invalid_json_alert: bool = False
        self._show_success_update: bool = False
        self._show_failure_update: bool = False
        self._show_outdated_configuration_warning: bool = False

    def get_associated_profile(self) -> AppProfile:
        return AppProfile.CONFIGURATION

    @classmethod
    def get_display_name(cls) -> str:
        return "Editar Configuración del Sistema"

    def _build_submenu_state(self) -> Dict:
        return {}

    def _build_content_state(self) -> Dict:
        return {}

    def _get_project_configuration(self):
        db = get_authenticated_client(
            self.app_state.authentication_info.role_service_account_credentials
        )
        self._project_configuration_data = get_current_project_model(db, self._tenant_id)
        self._current_gateway = list(
            self._project_configuration_data.gateways_table.keys()
        )[0]

    def _on_job_json_change(self):
        json_str = st.session_state[self._text_area_unique_key]
        json_dict = json.loads(json_str)
        job_config = JobConfig.build_from_dict(json_dict)

        if not job_config:
            self._show_invalid_json_alert = True
            log.error("Could not parse the edited Job configuration JSON")
        else:
            self._show_invalid_json_alert = False

        log.info("Job Config changed. Updating local information")
        self._project_configuration_data.jobs_table_by_gateway[
            self._current_gateway
        ] = job_config

    def on_gateway_selection_change(self):
        self._current_gateway = st.session_state[self._gateway_selectbox_unique_key]

    def on_create_snapshot_click(self, snapshot_type: ConfigurationSnapshotType):
        db = get_authenticated_client(
            self.app_state.authentication_info.role_service_account_credentials
        )

        # Applies a sentinel to see if the local state is outdated
        current_cloud_configuration = pull_current_configuration_from_firestore(db, self._tenant_id)
        current_cloud_configuration_version = current_cloud_configuration.to_dict()["configuration_id"]

        if current_cloud_configuration_version != self._project_configuration_data.configuration_version:
            log.warning("Trying to update cloud configuration but the local App data is outdated")
            self._show_outdated_configuration_warning = True
            db.close()
            return

        updated = create_configuration_snapshot(
            db, self._project_configuration_data, snapshot_type, self._tenant_id
        )
        db.close()
        if updated:
            self._show_success_update = True
            self._get_project_configuration()
        else:
            self._show_failure_update = True

    def _render_content(self) -> None:
        if not self._project_configuration_data:
            self._get_project_configuration()

        if self._show_invalid_json_alert:
            st.warning(
                "No se pudo parsear la configuración introducida. Intente nuevamente"
            )

        st.subheader(f"Configuración de trabajo {self._current_gateway}")

        if self._show_outdated_configuration_warning:
            st.warning("No se pudo actualizar la configuración porque necesita "
                       "descargar la configuración más reciente. Recargue esta ventana")
            self._show_outdated_configuration_warning = False

        if self._show_success_update:
            st.success("Se aplicó con éxito su actualización")
            self._show_success_update = False

        elif self._show_failure_update:
            st.error("Hubo un error y su configuración no se pudo aplicar")
            self._show_failure_update = False

        st.markdown(
            f"**Versión de configuración:** {self._project_configuration_data.configuration_version}"
        )

        gateway_job_config = self._project_configuration_data.jobs_table_by_gateway[
            self._current_gateway
        ]
        job_json = json.dumps(gateway_job_config, cls=ComplexEncoder, indent=4)
        st.text_area(
            f"Editar configuración",
            value=job_json,
            key=self._text_area_unique_key,
            height=600,
            on_change=self._on_job_json_change,
        )

        st.button(
            label="Aplicar version preliminar",
            help="Aplicar configuración preliminar con version X.X",
            on_click=self.on_create_snapshot_click,
            args=(ConfigurationSnapshotType.PRELIMINARY,),
        )

        st.button(
            label="Aplicar version",
            help="Aplicar configuración con version X",
            on_click=self.on_create_snapshot_click,
            args=(ConfigurationSnapshotType.NORMAL,),
        )

    def _render_submenu(self) -> None:
        if not self._project_configuration_data:
            self._get_project_configuration()

        st.markdown("**Actualizar configuración de trabajo**")
        st.selectbox(
            label="Seleccione el Gateway",
            options=list(self._project_configuration_data.gateways_table.keys()),
            key=self._gateway_selectbox_unique_key,
            on_change=self.on_gateway_selection_change,
        )



import numpy as np
from common import iCOMOX_messages, messages_utils
from scipy.signal import welch
from scipy.signal.windows import hamming
import struct


class class_ADXL1002:
    def __init__(self):
        # Initializing Fs_Acc, MFFT_Acc, window
        Root_clk = 26e6
        ACLKDIV = 4
        SAMPLETIME = 51
        OverSMP = 4
        DLY = 0

        if DLY == 0:
            self.Fs_Acc = (Root_clk / ACLKDIV) / ((14 + SAMPLETIME) * OverSMP)
        else:
            self.Fs_Acc = (Root_clk / ACLKDIV) / ((14 + SAMPLETIME) * OverSMP + DLY + 2)
        self.NFFT_Acc = 512 #iCOMOX_messages.ADXL1002_SAMPLES_NUM_PER_PIN    # 512
        self.Sensitivity_Acc = 0.9/50 # = 0.9[V]/50[g] = 0.018 [V/g] = 18 [mV/g]
        self.window = hamming(M=self.NFFT_Acc)

    def ADC_readings_to_g_units(self, ADC_samples):
        # convert ADC samples to g units. It is based on the following equation: V_ADC = 1.8 * ADC / 4096 and also V_ADC = 0.9 - Sensitivity_Acc * a = 0.9 - 0.018 * a
        # so we get that a[g] = 50 - 25*ADC/1024
        # The actual formula should be written with + instead of -, due to the inverting OPA:
        # a[g] = -50 + 25*ADC/1024
        samples = -50 + 25*np.array(ADC_samples)/1024   #(0.9/self.Sensitivity_Acc) * (1 - np.array(ADC_samples)/2048)
        # samples -= np.mean(samples)
        return  samples

    def g_units_to_PSD(self, acc_units_g):
        (freq_Acc, Pxxf_Acc) = welch(x=acc_units_g, fs=self.Fs_Acc, window=self.window, noverlap=self.NFFT_Acc/2, nfft=self.NFFT_Acc, return_onesided=True, detrend=False)
        return freq_Acc, Pxxf_Acc

    def msg_bytes_to_samples_g(self, payload):
        unpacked_bytes = messages_utils.ADC_Unpack_12bits_Array(packed_bytes=payload)
        ADC_samples = np.array(struct.unpack("<{}h".format(iCOMOX_messages.ADXL1002_SAMPLES_NUM), unpacked_bytes))
        ADC_samples = self.ADC_readings_to_g_units(ADC_samples=ADC_samples)
        return ADC_samples



from txp.web.resources.resources_management import ImageManagement
import streamlit as st
import dataclasses
import json

@dataclasses.dataclass
class ConditionReport:
    """A class to handle the asset condition report"""
    def __post_init__(self):
        self.conditions_mapping = {
            "OPTIMAL": self._optimal_condition,
            "GOOD": self._good_condition,
            "UNDEFINED": self._critical_condition,
            "OPERATIONAL": self._operational_condition,
            "CRITICAL": self._critical_condition,
        }

    @property
    def _critical_condition(self):
        return self._image_html_and_text("unstable_condition.jpg", "Mala", width=30)

    @property
    def _operational_condition(self):
        return self._image_html_and_text("operational_condition.jpg", "Operativa", width=30)

    @property
    def _optimal_condition(self):
        return self._image_html_and_text("optimal_condition.jpg", "Óptima", width=30)

    @property
    def _good_condition(self):
        return self._image_html_and_text("good_condition.jpg", "Buena", width=30)

    def _image_html(self, file_name, responsive=False, width=60):
        """
        returns an image html tag from a file_name
        """
        if responsive:
            resources = ImageManagement(file_name)
            html_responsive_resource = resources.get_local_html_responsive_resource(width)
            return html_responsive_resource
        else:
            resources = ImageManagement(file_name)
            html_resource = resources.get_local_html_resource(width)
            return html_resource

    def _image_html_and_text(self, file_name, text, width=60):
        """
        returns an image html tag from a file_name and includes the possibility of add text
        """
        img = self._image_html(file_name, width=width)
        return img + f"""<br>{text}"""


@dataclasses.dataclass
class EventReport:
    """A class to handle the event type report"""
    def __init__(self):
        self.events_mapping = self._built_report_to_events()

    @property
    def _detrimental(self):
        return self._image_html_and_text("detrimental_event.gif", "Perjudicial", width=30)

    @property
    def _inofensive(self):
        return self._image_html_and_text("inofensive_events.gif", "Inofensivo", width=30)

    def _image_html(self, file_name, responsive=False, width=60):
        """
        returns an image html tag from a file_name
        """
        if responsive:
            resources = ImageManagement(file_name)
            html_responsive_resource = resources.get_local_html_responsive_resource(width)
            return html_responsive_resource
        else:
            resources = ImageManagement(file_name)
            html_resource = resources.get_local_html_resource(width)
            return html_resource

    def _image_html_and_text(self, file_name, text, width=60):
        """
        returns an image html tag from a file_name and includes the possibility of add text
        """
        img = self._image_html(file_name, width=width)
        return img + f"""<br>{text}"""

    def _built_report_to_events(self):
        report = {}
        declarative_events= json.loads(st.secrets['event_type_report'])
        for event, value in declarative_events.items():
            if value == 'inofensive':
                report[event]=self._inofensive
            else:
                report[event]=self._detrimental
        return report



#!/usr/bin/env python3
#coding: utf-8
from sys import platform
import time

try:
    if platform == "Darwin" or platform == "darwin":
        import smbus2 as smbus
    else:
        import smbus
except:
    print('Warning: smbus library not installed. Maybe running on CI environment?')

class Arm_Device(object):

    def __init__(self):
        self.addr = 0x15
        self.bus = smbus.SMBus(1)


    def Arm_serial_servo_write(self, id, angle, time):
        if id == 0: 
            self.Arm_serial_servo_write6(angle, angle, angle, angle, angle, angle, time)
        elif id == 2 or id == 3 or id == 4:  
            angle = 180 - angle
            pos = int((3100 - 900) * (angle - 0) / (180 - 0) + 900)
            # pos = ((pos << 8) & 0xff00) | ((pos >> 8) & 0xff)
            value_H = (pos >> 8) & 0xFF
            value_L = pos & 0xFF
            time_H = (time >> 8) & 0xFF
            time_L = time & 0xFF
            try:
                self.bus.write_i2c_block_data(self.addr, 0x10 + id, [value_H, value_L, time_H, time_L])
            except:
                print('Arm_serial_servo_write I2C error')
        elif id == 5:
            pos = int((3700 - 380) * (angle - 0) / (270 - 0) + 380)
            # pos = ((pos << 8) & 0xff00) | ((pos >> 8) & 0xff)
            value_H = (pos >> 8) & 0xFF
            value_L = pos & 0xFF
            time_H = (time >> 8) & 0xFF
            time_L = time & 0xFF
            try:
                self.bus.write_i2c_block_data(self.addr, 0x10 + id, [value_H, value_L, time_H, time_L])
            except:
                print('Arm_serial_servo_write I2C error')
        else:
            pos = int((3100 - 900) * (angle - 0) / (180 - 0) + 900)
            # pos = ((pos << 8) & 0xff00) | ((pos >> 8) & 0xff)
            value_H = (pos >> 8) & 0xFF
            value_L = pos & 0xFF
            time_H = (time >> 8) & 0xFF
            time_L = time & 0xFF
            try:
                self.bus.write_i2c_block_data(self.addr, 0x10 + id, [value_H, value_L, time_H, time_L])
            except:
                print('Arm_serial_servo_write I2C error')


    def Arm_serial_servo_write_any(self, id, angle, time):
        if id != 0:
            pos = int((3100 - 900) * (angle - 0) / (180 - 0) + 900)
            value_H = (pos >> 8) & 0xFF
            value_L = pos & 0xFF
            time_H = (time >> 8) & 0xFF
            time_L = time & 0xFF
            try:
                self.bus.write_i2c_block_data(self.addr, 0x19, [id & 0xff, value_H, value_L, time_H, time_L])
            except:
                print('Arm_serial_servo_write_any I2C error')
        elif id == 0:  
            pos = int((3100 - 900) * (angle - 0) / (180 - 0) + 900)
            # pos = ((pos << 8) & 0xff00) | ((pos >> 8) & 0xff)
            value_H = (pos >> 8) & 0xFF
            value_L = pos & 0xFF
            time_H = (time >> 8) & 0xFF
            time_L = time & 0xFF
            try:
                self.bus.write_i2c_block_data(self.addr, 0x17, [value_H, value_L, time_H, time_L])
            except:
                print('Arm_serial_servo_write_any I2C error')


    def Arm_serial_servo_write_offset_switch(self, id):
        try:
            if id > 0 and id < 7:
                self.bus.write_byte_data(self.addr, 0x1c, id)
            elif id == 0:
                self.bus.write_byte_data(self.addr, 0x1c, 0x00)
                time.sleep(.5)
        except:
            print('Arm_serial_servo_write_offset_switch I2C error')


    def Arm_serial_servo_write_offset_state(self):
        try:
            self.bus.write_byte_data(self.addr, 0x1b, 0x01)
            time.sleep(.001)
            state = self.bus.read_byte_data(self.addr, 0x1b)
            return state
        except:
            print('Arm_serial_servo_write_offset_state I2C error')
        return None


    def Arm_serial_servo_write6_array(self, joints, time):
        s1, s2, s3, s4, s5, s6 = joints[0], joints[1], joints[2], joints[3], joints[4], joints[5]
        if s1 > 180 or s2 > 180 or s3 > 180 or s4 > 180 or s5 > 270 or s6 > 180:
            print("Parameter in range not 0-180！")
            return
        try:
            pos = int((3100 - 900) * (s1 - 0) / (180 - 0) + 900)
            value1_H = (pos >> 8) & 0xFF
            value1_L = pos & 0xFF

            s2 = 180 - s2
            pos = int((3100 - 900) * (s2 - 0) / (180 - 0) + 900)
            value2_H = (pos >> 8) & 0xFF
            value2_L = pos & 0xFF

            s3 = 180 - s3
            pos = int((3100 - 900) * (s3 - 0) / (180 - 0) + 900)
            value3_H = (pos >> 8) & 0xFF
            value3_L = pos & 0xFF

            s4 = 180 - s4
            pos = int((3100 - 900) * (s4 - 0) / (180 - 0) + 900)
            value4_H = (pos >> 8) & 0xFF
            value4_L = pos & 0xFF

            pos = int((3700 - 380) * (s5 - 0) / (270 - 0) + 380)
            value5_H = (pos >> 8) & 0xFF
            value5_L = pos & 0xFF

            pos = int((3100 - 900) * (s6 - 0) / (180 - 0) + 900)
            value6_H = (pos >> 8) & 0xFF
            value6_L = pos & 0xFF
            time_H = (time >> 8) & 0xFF
            time_L = time & 0xFF

            data = [value1_H, value1_L, value2_H, value2_L, value3_H, value3_L,
                    value4_H, value4_L, value5_H, value5_L, value6_H, value6_L]
            timeArr = [time_H, time_L]
            s_id = 0x1d
            self.bus.write_i2c_block_data(self.addr, 0x1e, timeArr)
            self.bus.write_i2c_block_data(self.addr, s_id, data)
        except:
            print('Arm_serial_servo_write6 I2C error')


    def Arm_serial_servo_write6(self, s1, s2, s3, s4, s5, s6, time):
        if s1 > 180 or s2 > 180 or s3 > 180 or s4 > 180 or s5 > 270 or s6 > 180:
            print("Parameter in range not 0-180！")
            return
        try:
            pos = int((3100 - 900) * (s1 - 0) / (180 - 0) + 900)
            value1_H = (pos >> 8) & 0xFF
            value1_L = pos & 0xFF

            s2 = 180 - s2
            pos = int((3100 - 900) * (s2 - 0) / (180 - 0) + 900)
            value2_H = (pos >> 8) & 0xFF
            value2_L = pos & 0xFF

            s3 = 180 - s3
            pos = int((3100 - 900) * (s3 - 0) / (180 - 0) + 900)
            value3_H = (pos >> 8) & 0xFF
            value3_L = pos & 0xFF

            s4 = 180 - s4
            pos = int((3100 - 900) * (s4 - 0) / (180 - 0) + 900)
            value4_H = (pos >> 8) & 0xFF
            value4_L = pos & 0xFF

            pos = int((3700 - 380) * (s5 - 0) / (270 - 0) + 380)
            value5_H = (pos >> 8) & 0xFF
            value5_L = pos & 0xFF

            pos = int((3100 - 900) * (s6 - 0) / (180 - 0) + 900)
            value6_H = (pos >> 8) & 0xFF
            value6_L = pos & 0xFF
            time_H = (time >> 8) & 0xFF
            time_L = time & 0xFF

            data = [value1_H, value1_L, value2_H, value2_L, value3_H, value3_L,
                    value4_H, value4_L, value5_H, value5_L, value6_H, value6_L]
            timeArr = [time_H, time_L]
            s_id = 0x1d
            self.bus.write_i2c_block_data(self.addr, 0x1e, timeArr)
            self.bus.write_i2c_block_data(self.addr, s_id, data)
        except:
            print('Arm_serial_servo_write6 I2C error')


    def Arm_serial_servo_read(self, id):
        if id < 1 or id > 6:
            print("id must be 1 - 6")
            return None
        try:
            self.bus.write_byte_data(self.addr, id + 0x30, 0x0)
            time.sleep(0.003)
            pos = self.bus.read_word_data(self.addr, id + 0x30)
        except:
            print('Arm_serial_servo_read I2C error')
            return None
        if pos == 0:
            return None
        pos = (pos >> 8 & 0xff) | (pos << 8 & 0xff00)
        # print(pos)
        if id == 5:
            pos = int((270 - 0) * (pos - 380) / (3700 - 380) + 0)
            if pos > 270 or pos < 0:
                return None
        else:
            pos = int((180 - 0) * (pos - 900) / (3100 - 900) + 0)
            if pos > 180 or pos < 0:
                return None
        if id == 2 or id == 3 or id == 4:
            pos = 180 - pos
        # print(pos)
        return pos


    def Arm_serial_servo_read_any(self, id):
        if id < 1 or id > 250:
            print("id must be 1 - 250")
            return None
        try:
            self.bus.write_byte_data(self.addr, 0x37, id)
            time.sleep(0.003)
            pos = self.bus.read_word_data(self.addr, 0x37)
        except:
            print('Arm_serial_servo_read_any I2C error')
            return None
        # print(pos)
        pos = (pos >> 8 & 0xff) | (pos << 8 & 0xff00)
        # print(pos)
        pos = int((180 - 0) * (pos - 900) / (3100 - 900) + 0)
        # print(pos)
        return pos


    def Arm_ping_servo(self, id):
        data = int(id)
        if data > 0 and data <= 250:
            reg = 0x38
            self.bus.write_byte_data(self.addr, reg, data)
            time.sleep(.003)
            value = self.bus.read_byte_data(self.addr, reg)
            times = 0
            while value == 0 and times < 5:
                self.bus.write_byte_data(self.addr, reg, data)
                time.sleep(.003)
                value = self.bus.read_byte_data(self.addr, reg)
                times += 1
                if times >= 5:
                    return None
            return value
        else:
            return None


    def Arm_get_hardversion(self):
        try:
            self.bus.write_byte_data(self.addr, 0x01, 0x01)
            time.sleep(.001)
            value = self.bus.read_byte_data(self.addr, 0x01)
        except:
            print('Arm_get_hardversion I2C error')
            return None
        version = str(0) + '.' + str(value)
        # print(version)
        return version


    def Arm_serial_set_torque(self, onoff):
        try:
            if onoff == 1:
                self.bus.write_byte_data(self.addr, 0x1A, 0x01)
            else:
                self.bus.write_byte_data(self.addr, 0x1A, 0x00)
        except:
            print('Arm_serial_set_torque I2C error')


    def Arm_serial_set_id(self, id):
        try:
            self.bus.write_byte_data(self.addr, 0x18, id & 0xff)
        except:
            print('Arm_serial_set_id I2C error')


    def Arm_Product_Select(self, index):
        try:
            self.bus.write_byte_data(self.addr, 0x04, index & 0xff)
        except:
            print('Arm_Product_Select I2C error')


    def Arm_RGB_set(self, red, green, blue):
        try:
            self.bus.write_i2c_block_data(self.addr, 0x02, [red & 0xff, green & 0xff, blue & 0xff])
        except:
            print('Arm_RGB_set I2C error')


    def Arm_Button_Mode(self, mode):
        try:
            self.bus.write_byte_data(self.addr, 0x03, mode & 0xff)
        except:
            print('Arm_Button_Mode I2C error')


    def Arm_reset(self):
        try:
            self.bus.write_byte_data(self.addr, 0x05, 0x01)
        except:
            print('Arm_reset I2C error')


    def Arm_PWM_servo_write(self, id, angle):
        try:
            if id == 0:
                self.bus.write_byte_data(self.addr, 0x57, angle & 0xff)
            else:
                self.bus.write_byte_data(self.addr, 0x50 + id, angle & 0xff)
        except:
            print('Arm_PWM_servo_write I2C error')


    def Arm_Clear_Action(self):
        try:
            self.bus.write_byte_data(self.addr, 0x23, 0x01)
            time.sleep(.5)
        except:
            print('Arm_Clear_Action I2C error')


    def Arm_Action_Study(self):
        try:
            self.bus.write_byte_data(self.addr, 0x24, 0x01)
        except:
            print('Arm_Action_Study I2C error')


    def Arm_Action_Mode(self, mode):
        try:
            self.bus.write_byte_data(self.addr, 0x20, mode & 0xff)
        except:
            print('Arm_Clear_Action I2C error')


    def Arm_Read_Action_Num(self):
        try:
            self.bus.write_byte_data(self.addr, 0x22, 0x01)
            time.sleep(.001)
            num = self.bus.read_byte_data(self.addr, 0x22)
            return num
        except:
            print('Arm_Read_Action_Num I2C error')


    def Arm_Buzzer_On(self, delay=0xff):
        try:
            if delay != 0:
                self.bus.write_byte_data(self.addr, 0x06, delay&0xff)
        except:
            print('Arm_Buzzer_On I2C error')


    def Arm_Buzzer_Off(self):
        try:
            self.bus.write_byte_data(self.addr, 0x06, 0x00)
        except:
            print('Arm_Buzzer_Off I2C error')

    def bus_servo_control(self, id, num, time=1000):
        try:
            # if num > 4000 or num < 96:
            #     print("bus_servo_control error, num must be [96, 4000]")
            #     return

            time_H = (time >> 8) & 0xFF
            time_L = time & 0xFF
        
            if id == 1 or id == 6:
                if num > 3100 or num < 900:
                    print("bus_servo_control error, num must be [900, 3100]")
                    return
                pos = int(num)
                value_H = (pos >> 8) & 0xFF
                value_L = pos & 0xFF
            elif id == 2 or id == 3 or id == 4:
                if num > 3100 or num < 900:
                    print("bus_servo_control error, num must be [900, 3100]")
                    return
                pos = int(3100 - num + 900)
                value_H = (pos >> 8) & 0xFF
                value_L = pos & 0xFF
            elif id == 5:
                if num > 4200 or num < 900:
                    print("bus_servo_control error, num must be [900, 4200]")
                    return
                pos = int(num - 514)
                value_H = (pos >> 8) & 0xFF
                value_L = pos & 0xFF
            else:
                print("bus_servo_control error, id must be [1, 6]")
                return
            self.bus.write_i2c_block_data(self.addr, 0x10 + id, [value_H, value_L, time_H, time_L])
        except:
            print('bus_servo_control error')

    def __change_value(self, value):
        try:
            val = 3100 - int(value) + 900
            return int(val)
        except:
            return None

    def bus_servo_control_array6(self, array, time=1000):
        try:
            if len(array) != 6:
                print("bus_servo_control_array6 input error")
                return

            s1, s2, s3, s4, s5, s6 = array[0], array[1], array[2], array[3], array[4], array[5]
            if s1 > 3100 or s2 > 3100 or s3 > 3100 or s4 > 3100 or s5 > 4200 or s6 > 3100:
                print("bus_servo_control_array6 input error")
                return
            elif s1 < 900 or s2 < 900 or s3 < 900 or s4 < 900 or s5 < 900 or s6 < 900:
                print("bus_servo_control_array6 input error")
                return

            pos = int(s1)
            value1_H = (pos >> 8) & 0xFF
            value1_L = pos & 0xFF

            s2 = self.__change_value(s2)
            pos = int(s2)
            value2_H = (pos >> 8) & 0xFF
            value2_L = pos & 0xFF

            s3 = self.__change_value(s3)
            pos = int(s3)
            value3_H = (pos >> 8) & 0xFF
            value3_L = pos & 0xFF

            s4 = self.__change_value(s4)
            pos = int(s4)
            value4_H = (pos >> 8) & 0xFF
            value4_L = pos & 0xFF

            s5 = s5 - 514
            pos = int(s5)
            value5_H = (pos >> 8) & 0xFF
            value5_L = pos & 0xFF

            pos = int(s6)
            value6_H = (pos >> 8) & 0xFF
            value6_L = pos & 0xFF

            time_H = (time >> 8) & 0xFF
            time_L = time & 0xFF

            data = [value1_H, value1_L, value2_H, value2_L, value3_H, value3_L,
                    value4_H, value4_L, value5_H, value5_L, value6_H, value6_L]
            timeArr = [time_H, time_L]
            s_id = 0x1d
            self.bus.write_i2c_block_data(self.addr, 0x1e, timeArr)
            self.bus.write_i2c_block_data(self.addr, s_id, data)
        except:
            print('bus_servo_control_array6 I2C error')
            



import threading
import serial
import time
import common.iCOMOX_messages
import common.helpers
import common
from common import helpers, iCOMOX_messages


class class_iCOMOX_Communication:
    def __init__(self, comm_port, callback_get_msg_size, callback_process_message, baudrate=115200, callback_open=None,
                 callback_close=None, duration_of_break_us=0, no_msg_read_timeout_sec=0,
                 callback_no_msg_read_timeout=None, serial_comm_thread=None):
        self.serObj = serial.Serial()
        self.serObj.baudrate = baudrate
        self.serObj.port = comm_port
        self.serObj.bytesize = serial.EIGHTBITS
        self.serObj.parity = serial.PARITY_NONE
        self.serObj.stopbits = serial.STOPBITS_TWO
        self.serObj.timeout = 1.0               # seconds, For read
        self.serObj.write_timeout = 1.0         # seconds
        self.serObj.interCharTimeout = None     # For both read & write
        self.serObj.xonxoff = False
        self.serObj.rtscts = False
        self.serObj.dsrdtr = False
        self.serObj.setRTS(False)
        self.serObj.setDTR(True)
        #self.serObj.set_buffer_size(rx_size=100000, tx_size=4096)
        self.Terminate = False      # Terminate is checked by the open/read thread
        self.no_msg_read_timeout_sec = no_msg_read_timeout_sec
        self.__abs_time_sec__ = time.time()
        self.duration_of_break_us = duration_of_break_us * 1E-6
        self.OnProcessMessages = callback_process_message
        self.OnGetIncomingMsgSize = callback_get_msg_size
        self.OnOpen = callback_open
        self.OnClose = callback_close
        self.OnNoMsgReadTimeout = callback_no_msg_read_timeout
        self.internal_receive_buffer = bytearray()  # it does not remain local variable for debug purpose
        self.received_msg = bytearray()             # it does not remain local variable for debug purpose
        if serial_comm_thread is None:
            self.thread = threading.Thread(target=self.__serial_comm_thread__)  # , args=(self,))
        else:
            self.thread = threading.Thread(target=serial_comm_thread)
        self.thread.start()

    def __delete__(self, instance):
        self.Terminate = True
        self.close()

    def test_for_no_msg_read_timeout(self):
        if (self.no_msg_read_timeout_sec != 0) and (time.time() - self.__abs_time_sec__ >= self.no_msg_read_timeout_sec):
            if callable(self.OnNoMsgReadTimeout):
                self.OnNoMsgReadTimeout(self)
            self.__abs_time_sec__ = time.time()
            return True
        else:
            return False

    def open(self, commPort=""):
        if self.serObj is None:
            return False
        if self.is_open():
            if self.serObj.port != commPort:
                self.serObj.close()
            else:
                return True
        try:
            self.serObj.port = commPort
            self.serObj.open()
        except Exception as ex:
            # print(ex.messaage)
            result = False
        else:
            result = True
        finally:
            if result and callable(self.OnOpen):
                self.OnOpen(comm=self)
            return result

    def is_open(self):
        return self.serObj.is_open

    def close(self):
        if self.is_open():
            self.serObj.flushOutput()
            self.serObj.close()
            if callable(self.OnClose):
                self.OnClose(comm=self)

#    def read(self, bytes_to_read):
#        result = bytearray()
#        try:
#            while (bytes_to_read > 0) and not self.Terminate:
#                bytes_read = self.serObj.read(bytes_to_read)
#                bytes_to_read -= len(bytes_read)
#                result += bytes_read
#                if self.test_for_no_msg_read_timeout():
#                    raise Exception("class_iCOMOX_Communication.read(): waiting too long without message been received")
#        except Exception as ex:     # Exception as ex:
#            result = bytearray()
#        finally:
#            return result

    def write(self, buffer_to_write):
        try:
            if self.duration_of_break_us != 0:
                self.serObj.send_break(self.duration_of_break_us)
            while not self.Terminate and (len(buffer_to_write) > 0):
                bytes_written = self.serObj.write(buffer_to_write)
                buffer_to_write = buffer_to_write[bytes_written:]

            result = len(buffer_to_write) == 0
        except:     # Exception as ex:
            result = False
        finally:
            return result

    #def find(self, arr, subarr, startpos=0):
    #    if (len(arr) <= startpos) or (len(subarr) == 0):
    #        return -1
    #    result = -1
    #    for result in range(startpos, len(arr)-len(subarr)+1):
    #        for j in range(0, len(subarr)):
    #            if arr[i+j] != subarr[j]:
    #                break
    # __serial_comm_thread__ is the thread function
    # It reads bytes from the port, accumulating them to meaningful messages by using OnGetIncomingMsgSize() and call
    #       OnProcessMessage() with the accumulated message
    # It can also signal the user if there is no activity for too long periods
    def __serial_comm_thread__(self):
        if not callable(self.OnGetIncomingMsgSize) or not callable(self.OnProcessMessages):
            raise Exception("communication.__serial_comm_thread__: OnGetIncomingMsgSize() and OnProcessMessages()"
                            " callbacks must be callable functions")
        print("SERIAL COMM THREAD INITIATED")
        # receive loop
        while not self.Terminate:
            # Wait until communication is open
            while not self.Terminate and not self.is_open():
                time.sleep(0.5)
            if self.Terminate:
                break

            # communication is open.
            self.__abs_time_sec__ = time.time()
            self.internal_receive_buffer = bytearray()
            self.received_msg = bytearray()
            header_received_flag = False
            while not self.Terminate and self.is_open():
                try:
                    self.internal_receive_buffer += self.serObj.read_all()
                except:
                    self.internal_receive_buffer = bytearray()
                    self.received_msg = bytearray()
                    self.close()
                    continue
                finally:
                    pass

                # phase 1: detect the header UART_IN_MSG_PREFIX
                if not header_received_flag:
                    header_pos = self.internal_receive_buffer.find(iCOMOX_messages.UART_IN_MSG_PREFIX, 0)
                    if header_pos >= 0:
                        self.internal_receive_buffer = self.internal_receive_buffer[header_pos+len(iCOMOX_messages.UART_IN_MSG_PREFIX):]
                        header_received_flag = True
                    else:
                        self.internal_receive_buffer = self.internal_receive_buffer[-len(iCOMOX_messages.UART_IN_MSG_PREFIX)+1:]
                else:
                    # phase 2: detect the message
                    adxl356_smip = False#common.app.Information.icomox_board_type == iCOMOX_messages.cCOMOX_BOARD_TYPE_SMIP
                    bytes_to_read = self.OnGetIncomingMsgSize(accumulated_msg=self.received_msg, adxl356_smip=adxl356_smip)
                    if bytes_to_read < 0:   # if invalid message
                        #helpers.OUT("Invalid message: {}".format(helpers.u8s_to_str(arr=self.received_msg, separator=" ", prefix="")))
                        break
                    elif bytes_to_read == 0:    # if complete receiving a valid message
                        self.OnProcessMessages(msg=self.received_msg, CommPort=self.serObj.port)
                        # if (len(self.received_msg) > 1) and (self.received_msg[0] == iCOMOX_messages.cCOMOX_MSG_CODE_Report):
                        #     helpers.OUT("Report message type 0x{:02X}, length {}".format(self.received_msg[1], len(self.received_msg)))
                        # else:
                        #     helpers.OUT("Received message type 0x{:02X}, length {}".format(self.received_msg[0], len(self.received_msg)))
                        self.received_msg = bytearray()
                        header_received_flag = False
                    else:                   # more bytes are needed to be accumulated for the current message
                        bytes_read = min(len(self.internal_receive_buffer), bytes_to_read)
                        self.received_msg += self.internal_receive_buffer[:bytes_read]
                        self.internal_receive_buffer = self.internal_receive_buffer[bytes_read:]
                self.test_for_no_msg_read_timeout()

        print("THREAD TERMINATED")
        # Thread was requested to terminate, then close the serial port
        self.close()
        helpers.OUT("Left __serial_comm_thread__()")



"""
    This script can be used to copy a collection in firestore from one document
    to another document.

    This is intended for manual operation of the manual_configuration_edit
    collection used for database initialization.
"""
import google.cloud.firestore as firestore


if __name__ == '__main__':
    collection_name = 'manual_configuration_edit'
    document_src = 'config_form'
    subcollection_name = 'jobs'  # Parse from input
    document_dst = 'gao_manual_script'

    db: firestore.Client = firestore.Client()

    # get collection
    # collection = db.collection(f'{collection_name}/{document_src}/{subcollection_name}').get()
    #
    # for doc in collection:
    #     db.collection(f'{collection_name}/{document_dst}/{subcollection_name}').document(doc.id).set(doc.to_dict())

    document = db.document('manual_configuration_edit/gao_manual_script/machines/gao_linea_de_pintura_asset')

    document.update(
        {
            'tasks': {
                'TP_01_SD': {
                    "task_id": "TP_01_SD",
                    "asset_id": "Linea_de_Pintura_GAO",
                    "task_type": "SlicPatchRecognitionTask",
                    "schema": {
                        'TXP-ARMROBOT-001_3e5be60c7a826a1e32eb9913ed11170f': {
                            "Image": {
                                'time': True
                            }
                        }
                    },  # Not relevant for tasks already trained
                    "edges": ['TXP-ARMROBOT-001_3e5be60c7a826a1e32eb9913ed11170f'],
                    "parameters": {
                        "number_of_segments": 40,
                        "image_shape": (640, 480, 3),
                        "thermal_image_shape": (120, 160, 3),
                        "patch_x_y_image": (100, 255),
                        "patch_x_y_thermal": (20, 8),
                        "pandas_cols": ['image', 'label']
                    }
                },
                'TP_02_MD': {
                    "task_id": "TP_02_MD",
                    "asset_id": "Linea_de_Pintura_GAO",
                    "task_type": "SlicPatchRecognitionTask",
                    "schema": {
                        'TXP-ARMROBOT-001_c0fcd732d555d7eb1752f65b83b8e6d5': {
                            "Image": {
                                'time': True
                            }
                        }
                    },  # Not relevant for tasks already trained
                    "edges": ['TXP-ARMROBOT-001_c0fcd732d555d7eb1752f65b83b8e6d5'],
                    "parameters": {
                        "number_of_segments": 40,
                        "image_shape": (640, 480, 3),
                        "thermal_image_shape": (120, 160, 3),
                        "patch_x_y_image": (200, 330),
                        "patch_x_y_thermal": (20, 8),
                        "pandas_cols": ['image', 'label']
                    }
                },
                'TP_03_ID': {
                    "task_id": "TP_03_ID",
                    "asset_id": "Linea_de_Pintura_GAO",
                    "task_type": "SlicPatchRecognitionTask",
                    "schema": {
                        'TXP-ARMROBOT-001_d5bdaeb38990169d310caa3e47b7a951': {
                            "Image": {
                                'time': True
                            }
                        }
                    },  # Not relevant for tasks already trained
                    "edges": ['TXP-ARMROBOT-001_d5bdaeb38990169d310caa3e47b7a951'],
                    "parameters": {
                        "number_of_segments": 40,
                        "image_shape": (640, 480, 3),
                        "thermal_image_shape": (120, 160, 3),
                        "patch_x_y_image": (220, 255),
                        "patch_x_y_thermal": (20, 8),
                        "pandas_cols": ['image', 'label']
                    }
                },
                'SD_01_SC': {
                    "task_id": "SD_01_SC",
                    "asset_id": "Linea_de_Pintura_GAO",
                    "task_type": "SlicPatchRecognitionTask",
                    "schema": {
                        'TXP-ARMROBOT-001_e135617e7c30e4fea19c15d764050d13': {
                            "Image": {
                                'time': True
                            }
                        }
                    },  # Not relevant for tasks already trained
                    "edges": ['TXP-ARMROBOT-001_e135617e7c30e4fea19c15d764050d13'],
                    "parameters": {
                        "number_of_segments": 40,
                        "image_shape": (640, 480, 3),
                        "thermal_image_shape": (120, 160, 3),
                        "patch_x_y_image": (220, 405),
                        "patch_x_y_thermal": (20, 8),
                        "pandas_cols": ['image', 'label']
                    }
                },
                'SD_02_MC': {
                    "task_id": "SD_02_MC",
                    "asset_id": "Linea_de_Pintura_GAO",
                    "task_type": "SlicPatchRecognitionTask",
                    "schema": {
                        'TXP-ARMROBOT-001_edc659ad2449fd102d70d1317446a437': {
                            "Image": {
                                'time': True
                            }
                        }
                    },  # Not relevant for tasks already trained
                    "edges": ['TXP-ARMROBOT-001_edc659ad2449fd102d70d1317446a437'],
                    "parameters": {
                        "number_of_segments": 40,
                        "image_shape": (640, 480, 3),
                        "thermal_image_shape": (120, 160, 3),
                        "patch_x_y_image": (240, 405),
                        "patch_x_y_thermal": (20, 8),
                        "pandas_cols": ['image', 'label']
                    }
                },
                'EMP_01_SI': {
                    "task_id": "EMP_01_SI",
                    "asset_id": "Linea_de_Pintura_GAO",
                    "task_type": "SlicPatchRecognitionTask",
                    "schema": {
                        'TXP-ARMROBOT-001_0a333974dfe6f2a6c55e8846ca5d82ff': {
                            "Image": {
                                'time': True
                            }
                        }
                    },  # Not relevant for tasks already trained
                    "edges": ['TXP-ARMROBOT-001_0a333974dfe6f2a6c55e8846ca5d82ff'],
                    "parameters": {
                        "number_of_segments": 40,
                        "image_shape": (640, 480, 3),
                        "thermal_image_shape": (120, 160, 3),
                        "patch_x_y_image": (70, 130),
                        "patch_x_y_thermal": (20, 8),
                        "pandas_cols": ['image', 'label']
                    }
                },
                'EMP_02_MI': {
                    "task_id": "EMP_02_MI",
                    "asset_id": "Linea_de_Pintura_GAO",
                    "task_type": "SlicPatchRecognitionTask",
                    "schema": {
                        'TXP-ARMROBOT-001_dfd684239c39d781de18751d9c62a72a': {
                            "Image": {
                                'time': True
                            }
                        }
                    },  # Not relevant for tasks already trained
                    "edges": ['TXP-ARMROBOT-001_dfd684239c39d781de18751d9c62a72a'],
                    "parameters": {
                        "number_of_segments": 40,
                        "image_shape": (640, 480, 3),
                        "thermal_image_shape": (120, 160, 3),
                        "patch_x_y_image": (150, 155),
                        "patch_x_y_thermal": (20, 8),
                        "pandas_cols": ['image', 'label']
                    }
                }
            }
        }
    )

    db.close()

    print("Elements successfully copied.")



from txp.ml.common.tasks import Policy
from txp.ml.common.tasks.task import Task, Encoder
import pandas as pd


class MockTask(Task):

    def __init__(self, task_definition=None, credentials_str=None):

        schema = None
        if task_definition is not None:
            schema = task_definition.schema_

        policy = MockPolicy()
        super().__init__(MockEncoder(schema), policy, task_definition, credentials_str)
        self.ready = True
        print("MockTask is ready.")

    def predict(self, sampling_window_rows) -> str:
        print(f"MockTask prediction request.")
        return self.task_definition.label_def.int_to_label[0]

    def train(self, dataset):
        self.policy.train(dataset, self.encoder.target)


class MockEncoder(Encoder):

    def __init__(self, schema):
        super().__init__(schema)

    def transform_input(self, signals_completed) -> pd.DataFrame:
        """
            At this moment we don't need to make predictions for Mock driver
        """
        return pd.DataFrame()

    def build_training_dataset(self, tables, target):
        """
            At this moment we don't need to make predictions for Mock driver
        """
        return pd.DataFrame()


class MockPolicy(Policy):

    def __init__(self, policy_file: str = None):
        super().__init__(policy_file)
        self.clf = None

    def name(self):
        return "MockPolicy"

    def train(self, dataset, target, test_size=0.3):
        """
            At this moment we don't need to make actual predictions
        """
        self.accuracy = 1
        self.trained = True

    def predict(self, tensor):
        """
           At this moment we don't need to make actual predictions
        """
        return None



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: gateway_config.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='gateway_config.proto',
  package='',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x14gateway_config.proto\x1a\x1cgoogle/protobuf/struct.proto\"\xd7\x01\n\x13SamplingWindowProto\x12\x18\n\x10observation_time\x18\x01 \x01(\x05\x12\x15\n\rsampling_time\x18\x02 \x01(\x05\x12\x14\n\x0cmachines_ids\x18\x03 \x03(\t\x12\x1d\n\x15observation_timestamp\x18\x04 \x01(\x03\x12\x17\n\x0fgateway_task_id\x18\x05 \x01(\x03\x12\x1d\n\x15sampling_window_index\x18\x06 \x01(\r\x12\"\n\x1anumber_of_sampling_windows\x18\x07 \x01(\r\"b\n\x14MachineMetadataProto\x12\x12\n\nmachine_id\x18\x01 \x01(\t\x12\r\n\x05model\x18\x02 \x01(\t\x12\x14\n\x0cmanufacturer\x18\x03 \x01(\t\x12\x11\n\tedges_ids\x18\x04 \x03(\t\"\x96\x01\n\x13\x45\x64geDescriptorProto\x12\x12\n\nlogical_id\x18\x01 \x01(\t\x12\x13\n\x0b\x64\x65vice_kind\x18\x02 \x01(\t\x12\x13\n\x0b\x64\x65vice_type\x18\x03 \x01(\t\x12\x31\n\x0b\x65\x64ge_params\x18\x04 \x01(\x0b\x32\x17.google.protobuf.StructH\x00\x88\x01\x01\x42\x0e\n\x0c_edge_params\"}\n\x12GatewayConfigProto\x12\x12\n\ngateway_id\x18\x01 \x01(\t\x12\x13\n\x0bregistry_id\x18\x02 \x01(\t\x12\x12\n\nproject_id\x18\x03 \x01(\t\x12\x14\n\x0c\x63loud_region\x18\x04 \x01(\t\x12\x14\n\x0cmachines_ids\x18\x06 \x03(\tb\x06proto3'
  ,
  dependencies=[google_dot_protobuf_dot_struct__pb2.DESCRIPTOR,])




_SAMPLINGWINDOWPROTO = _descriptor.Descriptor(
  name='SamplingWindowProto',
  full_name='SamplingWindowProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='observation_time', full_name='SamplingWindowProto.observation_time', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sampling_time', full_name='SamplingWindowProto.sampling_time', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='machines_ids', full_name='SamplingWindowProto.machines_ids', index=2,
      number=3, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='observation_timestamp', full_name='SamplingWindowProto.observation_timestamp', index=3,
      number=4, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='gateway_task_id', full_name='SamplingWindowProto.gateway_task_id', index=4,
      number=5, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sampling_window_index', full_name='SamplingWindowProto.sampling_window_index', index=5,
      number=6, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='number_of_sampling_windows', full_name='SamplingWindowProto.number_of_sampling_windows', index=6,
      number=7, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=55,
  serialized_end=270,
)


_MACHINEMETADATAPROTO = _descriptor.Descriptor(
  name='MachineMetadataProto',
  full_name='MachineMetadataProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='machine_id', full_name='MachineMetadataProto.machine_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='model', full_name='MachineMetadataProto.model', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='manufacturer', full_name='MachineMetadataProto.manufacturer', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='edges_ids', full_name='MachineMetadataProto.edges_ids', index=3,
      number=4, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=272,
  serialized_end=370,
)


_EDGEDESCRIPTORPROTO = _descriptor.Descriptor(
  name='EdgeDescriptorProto',
  full_name='EdgeDescriptorProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='logical_id', full_name='EdgeDescriptorProto.logical_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='device_kind', full_name='EdgeDescriptorProto.device_kind', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='device_type', full_name='EdgeDescriptorProto.device_type', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='edge_params', full_name='EdgeDescriptorProto.edge_params', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='_edge_params', full_name='EdgeDescriptorProto._edge_params',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=373,
  serialized_end=523,
)


_GATEWAYCONFIGPROTO = _descriptor.Descriptor(
  name='GatewayConfigProto',
  full_name='GatewayConfigProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='gateway_id', full_name='GatewayConfigProto.gateway_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='registry_id', full_name='GatewayConfigProto.registry_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='project_id', full_name='GatewayConfigProto.project_id', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='cloud_region', full_name='GatewayConfigProto.cloud_region', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='machines_ids', full_name='GatewayConfigProto.machines_ids', index=4,
      number=6, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=525,
  serialized_end=650,
)

_EDGEDESCRIPTORPROTO.fields_by_name['edge_params'].message_type = google_dot_protobuf_dot_struct__pb2._STRUCT
_EDGEDESCRIPTORPROTO.oneofs_by_name['_edge_params'].fields.append(
  _EDGEDESCRIPTORPROTO.fields_by_name['edge_params'])
_EDGEDESCRIPTORPROTO.fields_by_name['edge_params'].containing_oneof = _EDGEDESCRIPTORPROTO.oneofs_by_name['_edge_params']
DESCRIPTOR.message_types_by_name['SamplingWindowProto'] = _SAMPLINGWINDOWPROTO
DESCRIPTOR.message_types_by_name['MachineMetadataProto'] = _MACHINEMETADATAPROTO
DESCRIPTOR.message_types_by_name['EdgeDescriptorProto'] = _EDGEDESCRIPTORPROTO
DESCRIPTOR.message_types_by_name['GatewayConfigProto'] = _GATEWAYCONFIGPROTO
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

SamplingWindowProto = _reflection.GeneratedProtocolMessageType('SamplingWindowProto', (_message.Message,), {
  'DESCRIPTOR' : _SAMPLINGWINDOWPROTO,
  '__module__' : 'gateway_config_pb2'
  # @@protoc_insertion_point(class_scope:SamplingWindowProto)
  })
_sym_db.RegisterMessage(SamplingWindowProto)

MachineMetadataProto = _reflection.GeneratedProtocolMessageType('MachineMetadataProto', (_message.Message,), {
  'DESCRIPTOR' : _MACHINEMETADATAPROTO,
  '__module__' : 'gateway_config_pb2'
  # @@protoc_insertion_point(class_scope:MachineMetadataProto)
  })
_sym_db.RegisterMessage(MachineMetadataProto)

EdgeDescriptorProto = _reflection.GeneratedProtocolMessageType('EdgeDescriptorProto', (_message.Message,), {
  'DESCRIPTOR' : _EDGEDESCRIPTORPROTO,
  '__module__' : 'gateway_config_pb2'
  # @@protoc_insertion_point(class_scope:EdgeDescriptorProto)
  })
_sym_db.RegisterMessage(EdgeDescriptorProto)

GatewayConfigProto = _reflection.GeneratedProtocolMessageType('GatewayConfigProto', (_message.Message,), {
  'DESCRIPTOR' : _GATEWAYCONFIGPROTO,
  '__module__' : 'gateway_config_pb2'
  # @@protoc_insertion_point(class_scope:GatewayConfigProto)
  })
_sym_db.RegisterMessage(GatewayConfigProto)


# @@protoc_insertion_point(module_scope)



import sys
import os
sys.path.insert(0, os.path.join(sys.path[0], '..'))

import time
import threading

from SmartMeshSDK                 import ApiException
from SmartMeshSDK.ApiDefinition   import IpMgrDefinition, \
                                         HartMgrDefinition

import logging
class NullHandler(logging.Handler):
    def emit(self, record):
        pass
log = logging.getLogger('LatencyCalculator')
log.setLevel(logging.ERROR)
log.addHandler(NullHandler())

class LatencyCalculator(threading.Thread) :
    '''
    \brief Packet latency calculator for the SmartMesh IP network.
    '''
    QUERY_PERIOD_S     = 10
    
    def __init__(self,apiDef,connector):
        
        # log
        log.debug('Creating object')
        
        # store params
        self.apiDef                         = apiDef
        self.connector                      = connector
        
        # variables
        self.datalock                       = threading.Lock()
        self.lockPcTime                     = None
        self.lockNetworkTime                = None
        self.numNetworkTimeAcquisitions     = 0
        self.keepRunning                    = True
        self.delay                          = 0
        
        # initialize the parent class
        threading.Thread.__init__(self)
        self.name            = "LatencyCalculator"
        
    def run(self):
        # log
        log.debug('Thread started')
        
        self.delay           = 0
        while self.keepRunning:
            
            # sleep for a second
            time.sleep(1)
            
            if not self.keepRunning:
                log.debug('thread quitting')
                return
            
            # update time match
            if self.delay==0:
                
                # match PC time to manager time
                self.datalock.acquire()
                self.lockPcTime = time.time()
                try:
                    res = self.connector.dn_getTime()
                except ApiException.ConnectionError:
                    self.keepRunning = False
                    output = 'Connector appears disconnected, quitting thread'
                    log.error(output)
                    print output
                except Exception as err:
                    output = "unexpected {0}: {1}".format(type(err),err)
                    log.critical(output)
                    print output
                else:
                    if   isinstance(self.apiDef,IpMgrDefinition.IpMgrDefinition):
                        self.lockNetworkTime = float(res.utcSecs)+(float(res.utcUsecs)/1000000.0)
                    elif isinstance(self.apiDef,HartMgrDefinition.HartMgrDefinition):
                        self.lockNetworkTime = res.utc_time
                    else:
                        output = "apiDef of type {0} unexpected".format(type(self.apiDef))
                        log.critical(output)
                        print output
                        raise SystemError(output)
                    
                    self.numNetworkTimeAcquisitions += 1
                    
                    # log
                    log.info('Acquired time: lockNetworkTime={0} lockPcTime={1}'.format(
                                    self.lockNetworkTime,
                                    self.lockPcTime))
                finally:
                    self.datalock.release()
            
            # increment delay
            self.delay      += 1
            if self.delay==self.QUERY_PERIOD_S:
                # reset delay
                self.delay   = 0
    
    #======================== public ==========================================
    
    def getLatency(self,tsTxNetwork,tsRxPc):
        '''
        \brief Calculate the latency of a given packet
        
        \param tsTxNetwork [in] The timestamp (in s) as read from the data
                                notification. This is the time, expressed in
                                network time, when the packet was generated.
        \param tsRxPc     [in]  The timestamp (in epoch) of reception of this
                                packet at the PC.
        \exception RuntimeError if lock not acquired enough times yet.
        \returns The latency of the packet, in seconds.
        '''
        
        # log
        log.debug('Calculating latency for tsTxNetwork={0} tsRxPc={1}'.format(
                        tsTxNetwork,
                        tsRxPc))
        
        if self.numNetworkTimeAcquisitions<2:
            raise RuntimeError("lock hasn't been acquired enough times")
        
        # convert tsTxNetwork to tsTxPc
        self.datalock.acquire()
        tsTxPc     = self.lockPcTime + (tsTxNetwork - self.lockNetworkTime)
        self.datalock.release()
        
        log.debug('tsTxPc={0}'.format(tsTxPc))
        
        # calculate the latency
        latency    = tsRxPc - tsTxPc
        
        log.debug('latency={0}'.format(latency))
        
        return latency
    
    def disconnect(self):
        '''
        \brief Disconnect from the mamager.
        
        Causes the thread to terminate.
        '''
        
        # log
        log.warning('disconnect function called')
        
        # cause the thread to disconnect
        self.keepRunning = False

    #======================== private ==========================================    
    


import sys
from streamlit import cli as stcli
import os


def run_st_app():
    dirname = os.path.dirname(__file__)
    filename = os.path.join(dirname, 'streamlit_app.py')

    sys.argv = ["streamlit", "run", filename]
    sys.exit(stcli.main())


if __name__ == '__main__':
    run_st_app()




import socket
import threading
import ifaddr
# import BG96_Encoder
import time
import select
import errno
import sys

from common import iCOMOX_list, iCOMOX_messages, messages_utils
from common.common import iCOMOXs

cTCP_STATE_DISCONNECTED             = 0
cTCP_STATE_LISTEN                   = 1
cTCP_STATE_CLIENT_DISCONNECTED      = 2
cTCP_STATE_CLIENT_CONNECTED         = 3
cTCP_STATE_iCOMOX_CONNECTED         = 4
cTCP_STATE_ZOMBIE_CLIENT_DETECTED   = 5

def GetIpAddresses():
    result = []
    adapters = ifaddr.get_adapters()
    for adapter in adapters:
        # print("IPs of network adapter " + adapter.nice_name)
        for ip in adapter.ips:
            if type(ip.ip) is tuple:
                pass    #result += [ip.ip[0]]   # Collect only IPv4
            else:
                result += [ip.ip]
    return result

class class_iCOMOX_TcpServer:
    def __init__(self, callback_process_message, callback_state_changed):
        self.server = None
        self.Terminated = False
        self.OnProcessMessages = callback_process_message
        self.OnStateChanged = callback_state_changed

    def get_all_iComox_connected_as_TcpIp_clients(self):
        return [*filter(lambda iComox : iComox.socket.fileno() > 0, [*filter(lambda iComox : iComox.Type == iCOMOX_list.cCLIENT_TYPE_TCPIP, iCOMOXs.list)])]

    def send_to_all_clients(self, msg, UniqueID=None):
        iComox_clients = self.get_all_iComox_connected_as_TcpIp_clients()
        if UniqueID is not None:
            iComox_clients = [*filter(lambda iComox : iComox.UniqueID() == UniqueID, iComox_clients)]
        for iComox in iComox_clients:
            iComox.transmit_buffer += msg

    def is_listening(self):
        return self.server is not None

    def is_open(self):
        return self.is_listening() and (len(self.get_all_iComox_connected_as_TcpIp_clients()) > 0)

    def state_changed(self, tcpState, iComox = None):
        if callable(self.OnStateChanged):
            self.OnStateChanged(tcpState=tcpState, iComox=iComox)

    def close_client(self, iComox):
        if (iComox is None) or (iComox.Type != iCOMOX_list.cCLIENT_TYPE_TCPIP) or (iComox.socket is None):
            return

        iComox_clients = [*filter(lambda iCmx : (iCmx.Type == iCOMOX_list.cCLIENT_TYPE_TCPIP) and (iCmx.UniqueID() is not None), iCOMOXs.list)]
        listed_in_iComox_clients = iComox in iComox_clients
        try:
            if listed_in_iComox_clients:
                self.state_changed(tcpState=cTCP_STATE_CLIENT_DISCONNECTED, iComox=iComox)
            iComox.socket.close()
            if iCOMOXs.current == iComox:
                iCOMOXs.current = None
            iCOMOXs.list.remove(iComox)
            del iComox
            if listed_in_iComox_clients and (len(iComox_clients) == 0):
                self.state_changed(tcpState=cTCP_STATE_LISTEN, iComox=None)
        except Exception as ex:
            pass
        finally:
            pass

    # def get_all_clients_sockets(self):
    #     return [*map(lambda iComox: iComox.socket, [*filter(lambda iComox: (iComox.Type == iCOMOX_list.cCLIENT_TYPE_TCPIP) and (iComox.socket is not None), common.iCOMOXs.list)])]

    def __non_blocking_server_thread__(self, host, port):
        if not callable(self.OnProcessMessages):
            raise Exception("class_iCOMOX_TcpServer.__non_blocking_server_thread__.OnProcessMessages is not callable")
        iComox = None
        try:
            #self.OnStateChanged(tcpState=cTCP_STATE_DISCONNECTED)
            if self.server is not None:
                self.server.close()
            self.close_all_clients()
            self.server = socket.socket(family=socket.AF_INET, type=socket.SOCK_STREAM, proto=socket.IPPROTO_TCP)
            self.server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            self.server.bind((host, port))
            self.server.listen()
            self.server.setblocking(False)
            self.Terminated = False
            self.state_changed(tcpState=cTCP_STATE_LISTEN, iComox=None)

            while not self.Terminated:
                # time.sleep(0.001)
                if self.server is None:
                    self.state_changed(tcpState=cTCP_STATE_DISCONNECTED, iComox=None)
                    return
                try:
                    socks_list = [*map(lambda iComox: iComox.socket, self.get_all_iComox_connected_as_TcpIp_clients())]
                    if not socks_list:
                        socks_list = [self.server]
                    else:
                        socks_list = [self.server, *socks_list]
                    # socks_list = [self.server, *filter(lambda iComox : iComox.socket, [*filter(lambda iComox : (iComox.Type == iCOMOX_list.cCLIENT_TYPE_TCPIP) and (iComox.socket is not None), common.iCOMOXs.list)])]
                    # socks_list = [self.server, *[cli.socket for cli in self.clients]]
                    readable, writeable, errored = select.select(socks_list, socks_list, socks_list, 0)
                    for sock in errored:
                        iComox = iCOMOXs.find_by_socket(socket=sock)
                        if iComox is None:
                            continue
                        if sock in readable:
                            readable.remove(sock)
                        if sock in writeable:
                            writeable.remove(sock)
                        self.close_client(iComox=iComox)

                    for sock in writeable:
                        iComox = iCOMOXs.find_by_socket(socket=sock)
                        if iComox is None:
                            continue
                        if len(iComox.transmit_buffer) > 0:
                            bytes_sent = iComox.socket.send(iComox.transmit_buffer)
                            if bytes_sent != 0:
                                iComox.transmit_buffer = iComox.transmit_buffer[bytes_sent:]
                            else:
                                if sock in readable:
                                    readable.remove(sock)
                                self.close_client(iComox=iComox)

                    for sock in readable:
                        if sock is self.server:  # New iComox socket tries to connect. Here the accept does not block
                            new_client_socket, [remoteAddress, remotePort] = self.server.accept()
                            #self.setTcpKeepalive(sock=new_client_socket, enable=True, keepaliveIntervalSec=10, keepaliveProbeCount=2) #keepaliveIntervalSec=10
                            new_iComox = iCOMOXs.add(Type=iCOMOX_list.cCLIENT_TYPE_TCPIP, remoteAddress=remoteAddress, remotePort=remotePort, socket=new_client_socket)
                            self.state_changed(tcpState=cTCP_STATE_CLIENT_CONNECTED, iComox=new_iComox)
                        else:   # iComox sent data
                            iComox = iCOMOXs.find_by_socket(socket=sock)
                            if iComox is None:
                                continue
                            adxl356_smip = (iComox.Hello is None) or (iComox.board_type == iCOMOX_messages.cCOMOX_BOARD_TYPE_SMIP)
                            bytes_read = sock.recv(16384)
                            if bytes_read:
                                iComox.receive_buffer += bytes_read
                                while len(iComox.receive_buffer) >= 0:
                                    bytes_to_read = messages_utils.on_get_in_message_size(accumulated_msg=iComox.accumulated_msg, adxl356_smip=adxl356_smip)
                                    if bytes_to_read == 0:
                                        self.OnProcessMessages(msg=iComox.accumulated_msg, iComox=iComox)
                                        iComox.accumulated_msg = bytearray()

                                    elif bytes_to_read < 0: # if illegal message then close the socket
                                        self.close_client(iComox=iComox)
                                        break

                                    else:   # bytes_to_read > 0
                                        if len(iComox.receive_buffer) == 0: # if currently can't read more bytes then break
                                            break
                                        bytes_read = iComox.receive_buffer[:bytes_to_read]
                                        iComox.receive_buffer = iComox.receive_buffer[len(bytes_read):]
                                        bytes_to_read -= len(bytes_read)
                                        iComox.accumulated_msg += bytes_read
                            else:
                                self.close_client(iComox=iComox)    # Detection of FIN signal
                except socket.error as e:
                    err = e.args[0]
                    if (err != errno.EAGAIN) and (err != errno.EWOULDBLOCK):
                        self.close_client(iComox=iComox)
                finally:
                    pass

        except socket.error as e:
            # err = e.args[0]
            # if err == errno.EAGAIN or err == errno.EWOULDBLOCK:
            #     continue
            print(str(e))
            self.close_all_clients()
            self.server.close()
            self.server = None
        finally:
            self.Terminated = True
            self.state_changed(tcpState=cTCP_STATE_DISCONNECTED, iComox=None)

    def close_all_clients(self):
        iComoxs = [*filter(lambda iComox : (iComox.Type == iCOMOX_list.cCLIENT_TYPE_TCPIP) and (iComox.socket is not None), iCOMOXs.list)]
        for iComox in iComoxs:
            self.close_client(iComox=iComox)

    def start(self, host, port):
        if self.server is not None:
            self.server.close()
        self.Terminated = False
        new_server_thread = threading.Thread(target=self.__non_blocking_server_thread__, args=(host, port,))
        new_server_thread.daemon = True # Exit the server thread when the main thread terminates
        new_server_thread.start()

    def shutdown(self):
        if self.server is not None:
            self.Terminated = True
            self.server.close()
            self.server = None
            for iComox in self.get_all_iComox_connected_as_TcpIp_clients():
                self.close_client(iComox=iComox)
            self.state_changed(tcpState=cTCP_STATE_DISCONNECTED, iComox=None)

    def close(self):
        self.shutdown()

    def setTcpKeepalive(self, sock, enable, keepaliveIntervalSec, keepaliveProbeCount):
        # keepalive idle is the time until the first transmission of the keepalive packet
        # Keepalive time is the duration between two keepalive transmissions in idle condition.
        # Keepalive interval is the duration between two successive keepalive retransmissions, if acknowledgement to the previous keepalive transmission is not received.
        # Keepalive retry is the number of retransmissions to be carried out before declaring that remote end is not available
        if enable:
            enable = 1
        else:
            enable = 0

        if sys.platform.startswith("win"):  # Windows
            TCP_KEEPCNT = 16
            # keepalive enable, keepalive idle in miliseconds, keepalive interval in miliseconds
            sock.ioctl(socket.SIO_KEEPALIVE_VALS, (enable, int(keepaliveIntervalSec * 1000), int(keepaliveIntervalSec * 1000)))
            sock.setsockopt(socket.IPPROTO_TCP, TCP_KEEPCNT, keepaliveProbeCount)

        elif sys.platform.startswith("darwin"): # macOS
            sock.setsockopt(socket.SOL_SOCKET,  socket.SO_KEEPALIVE, enable)                  # on
            sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPALIVE, keepaliveIntervalSec)   # interval in seconds

        elif sys.platform.startswith("linux"):  # Linux
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, enable)                   # on
            sock.setsockopt(socket.SOL_TCP, socket.TCP_KEEPIDLE, keepaliveIntervalSec)        # interval in seconds
            sock.setsockopt(socket.SOL_TCP, socket.TCP_KEEPINTVL, keepaliveIntervalSec)       # interval in seconds
            sock.setsockopt(socket.SOL_TCP, socket.TCP_KEEPCNT, keepaliveProbeCount)
        else:
            raise Exception("Please add the set")



"""
This module defines classes to be used in the IPC between the Gateway
process and the UI process.

- UxClient: A UxClient class to be used by the Gateway instance in order to start
the streamlit application process and to send messages to it.

- _RemoteGatewayClient: a class to be used as singleton in the streamlit application
process. It will receive/send inter process messages from/to the gateway.
"""
# ============================ imports =========================================
from multiprocessing import Pipe
from multiprocessing.connection import Connection
from txp.devices.ux.streamlit_runner import StreamlitRunner
import time
from txp.common.config import settings
import logging

log = logging.getLogger(__name__)
log.setLevel(settings.txp.general_log_level)


# ==============================================================================
# Definition of the UxClient class.
# ==============================================================================
class UxClient:
    """The UxClient class provides a client for the gateway to start
    the streamlit app and perform IPC.

    The IPC is performed using Pipes. This allow for a two-way
    communication capability with a high level of abstraction:
        - Gateway can send and receive objects to/from the UX process.
        - Ux process can send and receive messages to/from the Gateway
        - The pipes are easily created: multiprocessing.Pipe()
    """

    def __init__(self):
        self._streamlit_runner: StreamlitRunner = None
        """_streamlit_runner: The StreamlitRunner process class to launch the 
        streamlit application in a separated process"""

        self._parent_connection: Connection
        """_parent_connection: Connection end to use in the Gateway process"""
        self._child_connection: Connection
        """_child_connection: Connection to use in the streamlit app process"""
        self._parent_connection, self._child_connection = Pipe()
        log.debug(
            f"UxClient: Created pipes connections. "
            f"Parent:{id(self._parent_connection)} - "
            f"Child:{id(self._child_connection)}"
        )

    def start_streamlit_app(self) -> None:
        """Starts the streamlit application programmatically.

        It pass down the end of the Pipe (_child_connection) to the StreamlitRunner.
        """
        kwargs = {"gateway_connection": self._child_connection}
        self._streamlit_runner = StreamlitRunner(kwargs)
        self._streamlit_runner.start()
        log.info("UxClient starting streamlit application in a separated process")

    def send_state_to_ui(self, ux_state) -> None:
        """Sends the appropriate UxState to the streamlit process
        based on the current Gateway state.

        Args:
            ux_state: An instance of a UxState defined in the ux.ux_states.py module.
        """
        log.debug(f"UxClient UxState sent to streamlit app.")
        self._parent_connection.send(ux_state)

    def wait_for_ux_state(self):
        """Returns the UXState received from the UI process.

        It'll block until there is something to receive.
        """
        while not self._parent_connection.poll():
            time.sleep(0.5)  # ask again in the next second.

        ux_state = self._parent_connection.recv()
        log.debug("UxClient UxState received from streamlit app")
        return ux_state




"""
    This script can be used to add task for machine in firestore.

    This is intended for manual operation of the manual_configuration_edit
    collection used for database initialization.
"""
import google.cloud.firestore as firestore

if __name__ == '__main__':
    collection_name = 'manual_configuration_edit'
    document_src = 'laboratorio_showroom_manual_script'
    subcollection_name = 'machines'
    machine_document = 'beer_demo'

    db: firestore.Client = firestore.Client()

    # get collection
    document = db.document(f'{collection_name}/{document_src}/{subcollection_name}/{machine_document}')

    document.update({
        'tasks': {
            'beer_spot_1': {
                'task_id': 'beer_spot_1',
                'asset_id': 'Showroom_Fridge',
                'task_type': 'SlicPatchRecognitionTask',
                'schema': {
                    'TXP-ARMROBOT-002_31378a021e68517f2e9b45948521ddbd': {
                        'Image': {
                            'time': True
                        }
                    }
                },
                'edges': ['TXP-ARMROBOT-002_31378a021e68517f2e9b45948521ddbd'],
                'parameters': {
                    'image_shape': [640, 480, 3],
                    'number_of_segments': 40,
                    'pandas_cols': ['image', 'label'],
                    'patch_x_y_image': [128, 82],
                    'patch_x_y_thermal': [20, 8],
                    'thermal_image_shape': [120, 160, 3]
                },
                'label_def': {
                    "labels": [
                        "PresentBeer",
                        "NotPresentBeer"
                    ],
                    "label_to_int": {
                        "PresentBeer": 1,  # Classification 1 True
                        "NotPresentBeer": 0  # Classification 0 False
                    },
                    "int_to_label": {
                        '1': "PresentBeer",
                        '0': "NotPresentBeer"
                    },
                    "label_to_taxonomy": {
                        "PresentBeer": "Harmless",
                        "NotPresentBeer": "Detrimental"
                    }
                }
            },
            'beer_spot_2': {
                'task_id': 'beer_spot_2',
                'asset_id': 'Showroom_Fridge',
                'task_type': 'SlicPatchRecognitionTask',
                'schema': {
                    'TXP-ARMROBOT-002_3271389c2310a31b4f893dbf33054a61': {
                        'Image': {
                            'time': True
                        }
                    }
                },
                'edges': ['TXP-ARMROBOT-002_3271389c2310a31b4f893dbf33054a61'],
                'parameters': {
                    'image_shape': [640, 480, 3],
                    'number_of_segments': 40,
                    'pandas_cols': ['image', 'label'],
                    'patch_x_y_image': [120, 60],
                    'patch_x_y_thermal': [20, 8],
                    'thermal_image_shape': [120, 160, 3]
                },
                'label_def': {
                    "labels": [
                        "PresentBeer",
                        "NotPresentBeer"
                    ],
                    "label_to_int": {
                        "PresentBeer": 1,  # Classification 1 True
                        "NotPresentBeer": 0  # Classification 0 False
                    },
                    "int_to_label": {
                        '1': "PresentBeer",
                        '0': "NotPresentBeer"
                    },
                    "label_to_taxonomy": {
                        "PresentBeer": "Harmless",
                        "NotPresentBeer": "Detrimental"
                    }
                }
            }
        }
    })




'''
Serial Mux Message classes
'''
import struct

# Magic token for the Serial Mux
MAGIC = struct.pack('BBBB', 0xa7, 0x40, 0xa0, 0xf5)

# Magic token for the Location Engine API
LOCATION_MAGIC = struct.pack('BBBB', 0x73, 0x03, 0xc6, 0x5d)

AUTH    = [ 48, 49, 50, 51, 52, 53, 54, 55 ]  # TODO: randomize me!
VERSION = 4

# Message Parser

class MuxMsg(object):
    def __init__(self, cb, ver = VERSION, magic = MAGIC, auth = AUTH):
        self.callback = cb
        self.ver = ver
        self.auth = auth
        self.magic = magic
        self.input_buffer = ''
    
    def getVer(self) :
        return self.ver
    
    def getAuth(self):
        return self.auth
    
    def build_message(self, cmd_type, data, cmd_id=0):
        '''Build a Serial Mux message'''
        cmd = self.magic
        cmd += struct.pack('!HHB', len(data)+3, cmd_id, cmd_type)
        cmd += data
        return cmd
    
    def parse(self, data):
        '''
        Parse input data
        Calls the registered callback when complete message is received
        '''
        if not data:
            return
        self.input_buffer += data
        while self.parse_one():
            pass

    def parse_one(self):
        '''Parse a single command from input_data
        Returns: whether a command was found
        '''
        msg_start = self.input_buffer.find(self.magic)
        if msg_start >= 0:
            # strip the ignored input
            self.input_buffer = self.input_buffer[msg_start:]
            # verify input is long enough
            if len(self.input_buffer) < 6:
                return False
            # parse message header
            bin_len = self.input_buffer[4:6]
            msg_len = struct.unpack('!H', bin_len)[0]
            # TODO: limit the length of valid messages
            index_end = 6 + msg_len
            # verify the message is complete
            if len(self.input_buffer) < index_end:
                return False
            
            msg = self.input_buffer[6:index_end]
            (cmd_id, cmd_type) = struct.unpack('!HB', msg[0:3])
            data = msg[3:]
            if self.callback:
                self.callback(cmd_id, cmd_type, data)
            self.input_buffer = self.input_buffer[index_end:]
            return True
        else:
            # if the token doesn't appear, ignore all but the last 3 characters
            self.input_buffer = self.input_buffer[-3:]
            return False
            
